#!/usr/bin/env python
import os
from .lib import Server
from .fetch import check_uri
from subprocess import run
import sys
import time
import os

DEFAULT_DOCKER_OPTION = ''
DEFAULT_SERVER = os.environ.get('SCITQ_SERVER','127.0.0.1')

def get_sample(filename):
    """Return last dir of a path"""
    return filename.split('/')[-1]

def check_unique(variable, value, default_value=None):
    "check if an option was not set twice"
    if value!=default_value:
        raise SyntaxError(f'{variable} option was set twice.')


def launch(docker, docker_option, uid, gid, 
            command, server, test, name, batch, input, output, resource):
    """A launching function for scitq tasks
    """

    s=Server(server)

    if docker is not None:
        #command = f'docker run --rm -u {uid}:{gid} {docker_option} {docker} {command}'
        if uid is not None and gid is not None:
            docker_option = f'-u {uid}:{gid} {docker_option}'
        else:
            docker_option = f'{docker_option}'
    elif uid == None and gid == None :
        command = f'{command}'
    else:
        command = f'sudo -u "#{uid}" -g "#{gid}" {command}'
    
    if test:
        print(f'Command that would be sent: {command}')
    else:
        s.task_create(command, name=name, batch=batch, container=docker,
            container_options=docker_option, input=input, output=output,
            resource=resource)

def command_join(command_list):
    """rejoin a shlex.splitted command
    shlex.join should do that but does strange things with some operands (<,>,|)"""
    return " ".join([word if ' ' not in word else (
            '"{}"'.format(word) if "'" in word else "'{}'".format(word) )
        for word in command_list])

def main():
    argv = list(sys.argv[1:])
    docker = None
    uid = None
    gid = None
    server = DEFAULT_SERVER
    docker_option = DEFAULT_DOCKER_OPTION
    test = False
    name = None
    batch = None
    input = ''
    output = None
    resource = ''
    try:
        while True:
            if len(argv)==0 or argv[0] in ['-h','--help']:
                raise SyntaxError('help')
            elif argv[0] in ['-d','--docker']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: DOCKERNAME')
                check_unique('docker',docker)
                docker = argv.pop(0)
            elif argv[0] in ['-O','--option']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: DOCKEROPTION')
                check_unique('docker_option',docker_option,DEFAULT_DOCKER_OPTION)
                docker_option = argv.pop(0)
            elif argv[0] in ['-u', '--uid']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: UID')
                check_unique('uid',uid)
                uid = int(argv.pop(0))
            elif argv[0] in ['-g', '--gid']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: GID')
                check_unique('gid',gid)
                gid = int(argv.pop(0))
            elif argv[0] in ['-s', '--server']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: scitq server name or IP')
                check_unique('server',server, DEFAULT_SERVER)
                server = argv.pop(0)
            elif argv[0] in ['-t', '--test']:
                option = argv.pop(0)
                test = True  
            elif argv[0] in ['-n', '--name']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: NAME, the name of the task')
                check_unique('name',name)
                name = argv.pop(0)
            elif argv[0] in ['-b', '--batch']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: BATCH, the batch of this task')
                check_unique('batch',batch)
                batch = argv.pop(0)
            elif argv[0] in ['-i', '--input']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: INPUT a quoted space-separated list of file URI')
                if input:
                    input = ' '.join([input,argv.pop(0)])
                else:
                    input = argv.pop(0)
            elif argv[0] in ['-r', '--resource']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: INPUT a quoted space-separated list of file URI')
                if resource:
                    resource = ' '.join([resource,argv.pop(0)])
                else:
                    resource = argv.pop(0)
            elif argv[0] in ['-o', '--output']:
                option = argv.pop(0)
                if len(argv)==0:
                    raise SyntaxError(f'{option} requires at least one argument: OUTPUT a unique directory URI')
                check_unique('output',output)
                output = argv.pop(0)
            elif argv[0]=='--':
                argv.pop(0)
                command = command_join(argv)
                break     
            else:
                command = command_join(argv)
                break
        if docker_option != DEFAULT_DOCKER_OPTION and docker is None:
            raise SyntaxError('Do not set Docker options if you do not specify a docker.')
        for category,items in (('input', input.split()), ('resource', resource.split()), ('output', [output])):
            for item in items:
                try:
                    if item is not None:
                        check_uri(item)
                except Exception as e:
                    raise SyntaxError(f'Defect in {category}: {e.args[0]}')
        launch(docker=docker, docker_option=docker_option, uid=uid, gid=gid,
            command=command, server=server, test=test, name=name, batch=batch,
            input=input, output=output, resource=resource)
    except SyntaxError as e:
        if e.args[0]!='help':
            print('Syntax error : {}'.format(e.args[0]), file=sys.stderr)
            sys.exit(2)
        print(f"""Usage:
scitq-launch     [(-h|--help)] [(-d|--docker) DOCKERNAME [(-o|--option) DOCKEROPTION]] 
                [(-u|--uid) UID] [(-g|--gid) GID] [(-s|--server) SERVER] [(-t|--test)] 
                [(-n|--name) NAME] [(-b|--batch) BATCH] [(-i|--input) INPUT] 
                [(-o|--output) OUTPUT] [(-r|--resource) OUTPUT] [--] COMMAND
    Add a task in scitq with :
    -h,--help       display this help message
    -d,--docker     execute the task in a docker container (named DOCKERNAME)
                    default to None - the task is executed natively
    -O,--option     (O is capital o) set docker options to DOCKEROPTION 
                    (default to '{DEFAULT_DOCKER_OPTION}') 
    -u,--uid        execute the task as this UID (with -u in docker or in sudo
                    if docker is not used). Default to current user UID ({uid})
                    BE CAREFUL: if not with docker, the corresponding user must 
                    exist on (remote) system,
    -g,--gid        execute the task as this GID (with -u in docker or -g in 
                    sudo). Default to current user GID ({gid})
                    BE CAREFUL: if not with docker, the corresponding group must 
                    1) exists on (remote) system,
                    2) user with uid UID must be a member of this group;
    -s,--server     send the order to server SERVER (default to {DEFAULT_SERVER})
                    (this default value can be changed using SCITQ_SERVER environment
                    variable)
    -t,--test       print the actual final task sent to the server instead
                    of really sending it (for debug purpose)
    -n,--name       set the name of the command, default to None, no name
    -b,--batch      set the name of the batch, default to None, default batch
    -i,--input      set some INPUT URIs (quoted and space separated) like 
                    s3://.../...fq.gz or ftp:///.../...fq.gz that must be preloaded
                    (in /input directory in docker, or in the directory specified
                    in INPUT environment variable if docker is not used)
                    NB **new** input option can be specified several times
    -o,--output     set OUTPUT, a directory URI in which all output result of the
                    task will be copied (output is simply the content of /output
                    directory in docker, or in the directory specified in OUTPUT
                    environment variable if docker is not used)
    -r,--resource   set some RESOURCE URIs (quoted and space separated) very much
                    like INPUT (see -i) except RESOURCEs are only downloaded once
                    per worker and shared between tasks (in /resource directory 
                    in docker or in the directory specified in RESOURCE environment
                    variable if docker is not used). It should be specified for 
                    all the tasks that need it but it will only be downloaded once
                    per host (provided the RESOURCE is specified exactly in the 
                    same way for each task). 
                    NB a RESOURCE URI may end with '|gunzip' in which case it
                    will be unzipped (with pigz) upon reception.
                    NB2 resource option can be specified several times
    --              the last argument, COMMAND, may be constituted of several
                    words. Usually this list of words starts with a non-dashed 
                    word like sh or bash, in which case there is no possible 
                    ambiguity. But in some case, the command may start with a 
                    dash in which case it may overlap with one the above option:
                    to avoid that, specifying -- will make sure the next word, 
                    whatever it is, is interpreted as the beginning of the 
                    command. 
    
    COMMAND is a usual shell command

    Be careful that if you do not define INPUT/OUTPUT and use docker (-d) then 
    all data transfer is manually done within your task. /data is generally used
    in that case, it is mounted in worker if an nfs_server variable is defined
    in ansible (for manager group, see /etc/ansible/inventory/common) - and NFS
    is minimally set up on that server. 
""")
        sys.exit(2)



if __name__=="__main__":
    main()



from argparse import Namespace
from datetime import datetime
from flask import Flask, render_template, jsonify, request
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import func, and_, select, delete, true, event, DDL
from sqlalchemy.orm import Session
from sqlalchemy.exc import ProgrammingError
from flask_restx import Api, Resource, fields
from werkzeug.middleware.proxy_fix import ProxyFix
from time import sleep
from threading import Thread
import queue
import logging as log
from logging.config import dictConfig
import os
from subprocess import run, Popen, PIPE
import signal
import json as json_module
from sqlalchemy.dialects import sqlite
from .util import PropagatingThread, package_path, package_version, check_dir, to_dict, tryupdate
from .default_settings import SQLALCHEMY_POOL_SIZE, SQLALCHEMY_DATABASE_URI
from .ansible.scitq.sqlite_inventory import scitq_inventory
from .constants import SIGNAL_CLEAN, SIGNAL_RESTART


MAIN_THREAD_SLEEP = 5
WORKER_OFFLINE_DELAY = 15
SCITQ_SERVER = os.environ.get('SCITQ_SERVER',None)

WORKER_CREATE = f'cd {package_path("ansible","playbooks")} && ansible-playbook deploy_one_vm.yaml --extra-vars \
"nodename={{hostname}} concurrency={{concurrency}} status=running flavor={{flavor}} \
region={{region}} provider={{provider}}"'

if SCITQ_SERVER is not None:
    WORKER_CREATE = WORKER_CREATE[:-1] + f' target={SCITQ_SERVER}"'
WORKER_DELETE = os.environ.get('WORKER_DELETE',
    f'cd {package_path("ansible","playbooks")} && ansible-playbook destroy_vm.yaml --extra-vars "nodename={{hostname}}"')
SERVER_CRASH_WORKER_RECOVERY = os.environ.get('SERVER_CRASH_WORKER_RECOVERY',
    f'cd {package_path("ansible","playbooks")} && ansible-playbook check_after_reboot.yaml')
WORKER_IDLE_CALLBACK = os.environ.get('WORKER_IDLE_CALLBACK',WORKER_DELETE)
WORKER_CREATE_CONCURRENCY = 10
WORKER_CREATE_RETRY=2
WORKER_CREATE_RETRY_SLEEP=30
UI_OUTPUT_TRUNC=100
UI_MAX_DISPLAYED_ROW = 500
WORKER_DESTROY_RETRY=2

if os.environ.get('QUEUE_PROCESS') and os.environ.get('QUEUE_LOG_FILE'):
    check_dir(os.environ.get('QUEUE_LOG_FILE'))
    dictConfig({
        'version': 1,
        'formatters': {'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }, "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "default",
            "filename": os.environ.get('QUEUE_LOG_FILE'),
            "maxBytes": int(os.environ.get('QUEUE_LOG_FILE_MAX_SIZE',
                os.environ.get('LOG_FILE_MAX_SIZE',"10000000"))),
            "backupCount": int(os.environ.get('QUEUE_LOG_FILE_KEEP',
                os.environ.get('LOG_FILE_KEEP',"3")))
        }},
        'root': {
            'level': os.environ.get('LOG_LEVEL',"INFO"),
            'handlers': ['wsgi' if 'DEBUG' in os.environ else 'file']
        }
    })
else:
    check_dir(os.environ.get('LOG_FILE',"/tmp/scitq.log"))
    dictConfig({
        'version': 1,
        'formatters': {'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }, "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "default",
            "filename": os.environ.get('LOG_FILE',"/tmp/scitq.log"),
            "maxBytes": int(os.environ.get('LOG_FILE_MAX_SIZE',"10000000")),
            "backupCount": int(os.environ.get('LOG_FILE_KEEP',"3"))
        }},
        'root': {
            'level': os.environ.get('LOG_LEVEL',"INFO"),
            'handlers': ['wsgi' if 'DEBUG' in os.environ else 'file']
        }
    })

IS_SQLITE = 'sqlite' in SQLALCHEMY_DATABASE_URI


log.info('Starting')
log.warning(f'WORKER_CREATE is {WORKER_CREATE}')

#worker_create_queue = queue.Queue()


# via https://github.com/pallets/flask-sqlalchemy/blob/main/examples/hello/hello.py
app = Flask(__name__, instance_relative_config=True)
#app.config['SQLALCHEMY_DATABASE_URI'] = "postgresql://postgres:postgres@localhost:5432/"
app.config.from_object('scitq.default_settings')
#app.config.from_pyfile("scitq.cfg", silent=True)
#app.config.from_prefixed_env()
if SQLALCHEMY_POOL_SIZE is not None:
    db = SQLAlchemy(app, engine_options={'pool_size': int(SQLALCHEMY_POOL_SIZE)})
else:
    db = SQLAlchemy(app)

# with uwsgi, the master worker is forking to create the workers which receive a 
# non-working connection (because it comes from another process), it must be discarded
# so that workers re-open the connection properly
# idea from https://stackoverflow.com/questions/39562838/how-to-configure-pyramid-uwsgi-sqlalchemy
try:
    # import uwsgi is only working in uwsgi context. It is normal that is fails
    # to import in VisualStudioCode or manually
    import uwsgi # pyright: ignore[reportMissingImports]

    def postfork():
        db.engine.dispose()
    uwsgi.post_fork_hook = postfork
except ImportError:
    pass

 #######  ######   ######  
 #     #  #     #  #     # 
 #     #  #     #  #     # 
 #     #  ######   ######  
 #     #  #   #    #     # 
 #     #  #    #   #     # 
 #######  #     #  ######  




class Task(db.Model):
    __tablename__ = "task"
    task_id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String)
    command = db.Column(db.String, nullable=False)
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    batch = db.Column(db.String, nullable=True)
    input = db.Column(db.String, nullable=True)
    output = db.Column(db.String, nullable=True)
    container = db.Column(db.String, nullable=True)
    container_options = db.Column(db.String, nullable=True)
    resource = db.Column(db.String, nullable=True)

    def __init__(self, command, name=None, status='pending', batch=None, 
                    input=None, output=None, container=None, 
                    container_options=None, resource=None):
        self.name = name
        self.command = command
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.batch = batch
        self.input = input
        self.output = output
        self.container = container
        self.container_options = container_options
        self.resource = resource


class Worker(db.Model):
    __tablename__ = "worker"
    worker_id = db.Column(db.Integer,primary_key=True)
    name = db.Column(db.String, nullable=False)
    hostname = db.Column(db.String)
    status = db.Column(db.String, nullable=False)
    concurrency = db.Column(db.Integer)
    prefetch = db.Column(db.Integer)
    load = db.Column(db.String)
    memory=db.Column(db.String)
    stats=db.Column(db.String)
    creation_date = db.Column(db.DateTime)
    modification_date = db.Column(db.DateTime)
    last_contact_date = db.Column(db.DateTime)
    batch = db.Column(db.String,nullable=True)
    idle_callback = db.Column(db.String, nullable=True)

    def __init__(self, name, concurrency, prefetch=0, hostname=None, 
                status='paused', batch=None, idle_callback=None):
        self.name = name
        self.concurrency = concurrency
        self.prefetch = prefetch
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.hostname = hostname
        self.batch = batch
        self.idle_callback = idle_callback
    

class Execution(db.Model):
    __tablename__ = "execution"
    execution_id = db.Column(db.Integer, primary_key=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"), nullable=True)
    worker = db.relationship(
        Worker,
        backref=db.backref('executions',
                         uselist=True,
                         cascade='save-update',
                         order_by='Execution.modification_date.desc()'))
    task_id = db.Column(db.Integer, db.ForeignKey("task.task_id"), nullable=False)
    task = db.relationship(
        Task,
        backref=db.backref('tasks',
                         uselist=True,
                         cascade='delete,all'),
                         order_by='Execution.creation_date')
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    output = db.Column(db.Text)
    error = db.Column(db.Text)
    return_code = db.Column(db.Integer)
    pid = db.Column(db.String)
    output_files = db.Column(db.String, nullable=True)
    command = db.Column(db.String, nullable=True)
    latest = db.Column(db.Boolean, default=True)
    

    def __init__(self, worker_id, task_id, status='pending', pid=None, 
                    return_code=None, command=None):
        self.worker_id = worker_id
        self.task_id = task_id
        self.status = status
        self.pid = pid
        self.return_code = return_code
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.command = command

trigger_latest_sqlite = DDL("""
        CREATE TRIGGER is_latest BEFORE INSERT ON execution FOR EACH ROW 
        BEGIN
            UPDATE execution SET latest=false WHERE latest AND task_id=NEW.task_id;
        END
        """)

func_latest_postgres = DDL("""
        CREATE OR REPLACE FUNCTION render_obsolete() 
        RETURNS TRIGGER
        LANGUAGE plpgsql
        AS $$
        BEGIN
            UPDATE execution 
            SET latest=false
            WHERE task_id=NEW.task_id
            AND latest;
            RETURN NEW;
        END;
        $$
        """)

trigger_latest_postgres = DDL("""
CREATE TRIGGER is_latest BEFORE INSERT ON execution FOR EACH ROW EXECUTE PROCEDURE render_obsolete()
""")

event.listen(
    Execution.__table__, 'after_create',
    trigger_latest_sqlite.execute_if(dialect="sqlite")    
)
event.listen(
    Execution.__table__, 'after_create',
    func_latest_postgres.execute_if(dialect="postgresql")    
)
event.listen(
    Execution.__table__, 'after_create',
    trigger_latest_postgres.execute_if(dialect="postgresql")    
)

class Signal(db.Model):
    __tablename__ = "signal"
    #execution_id = db.Column(db.Integer, db.ForeignKey("execution.execution_id"), primary_key=True, nullable=True)
    signal_id = db.Column(db.Integer, primary_key=True)
    execution_id = db.Column(db.Integer, db.ForeignKey("execution.execution_id"), nullable=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"))
    signal = db.Column(db.Integer, nullable=False)    

    def __init__(self, execution_id, worker_id, signal):
        self.execution_id = execution_id
        self.worker_id = worker_id
        self.signal = signal


class Job(db.Model):
    __tablename__ = "job"
    job_id = db.Column(db.Integer, primary_key=True)
    target = db.Column(db.String)
    action = db.Column(db.String)
    args = db.Column(db.JSON, default={})
    retry =  db.Column(db.Integer, default = 0)
    status = db.Column(db.String, default = 'pending')
    log = db.Column(db.Text, nullable=True)
    creation_date = db.Column(db.DateTime, server_default=func.now())
    modification_date = db.Column(db.DateTime, onupdate=func.now())
    

    def __init__(self, target, action, args={}, retry=0):
        self.target = target
        self.action = action
        self.args = args
        self.retry = retry


def create_worker_destroy_job(worker, session, commit=True):
    job = Job(target = worker.name,
        action='worker_destroy',
        args=to_dict(worker),
        retry=WORKER_DESTROY_RETRY)
    session.add(job)
    if commit:
        session.commit()


with app.app_context():
    db.create_all()


 ######      #     ####### 
 #     #    # #    #     # 
 #     #   #   #   #     # 
 #     #  #     #  #     # 
 #     #  #######  #     # 
 #     #  #     #  #     # 
 ######   #     #  ####### 
                           


def process_filtering_args(ObjectType, args):
    """In list functions it seems natural to filter the results with some attributes
    conditions (such as 'batch="foo"' or 'status="succeeded"') and it makes more 
    efficient queries than listing the lot and filtering the results out or sending
    several individual queries (one per item id).
    However sqlalchemy filtering syntax is not friendly when using typical python
    optional args dictionary, this function does this transformation:
    
    When args maybe something as 'batch="foo"' or 'batch=["foo","bar"]'
    - the function will return 'ObjectType.batch=="foo"'
    - or 'ObjectType.batch.in_(["foo","bar"])' if value is a list

    which can then be passed to sqlalchemy .filter() function
    """
    # 
    # filter expects something as 'ObjectType.batch=="foo"'
    # or 'ObjectType.batch.in_(["foo","bar"])' if value is a list
    new_args = [ getattr(ObjectType,attribute).in_(value) if type(value)==list else \
                 getattr(ObjectType,attribute)==value 
                    for attribute,value in args.items() ]
    return and_(*new_args) if len(new_args)>1 else new_args[0]


# via https://flask-restx.readthedocs.io/en/latest/example.html
app.wsgi_app = ProxyFix(app.wsgi_app)
api = Api(app, version='1.0', title='TaskMVC API',
    description='A simple TaskMVC API'
)


class BaseDAO(object):
    ObjectType = None
    authorized_status = []

    def get(self, id):
        object = self.ObjectType.query.get(id)
        if object is None:
            api.abort(404, "{} {} doesn't exist".format(
                    self.ObjectType.__name__,id))
        return object
        
    def create(self, data):
        object = self.ObjectType(**data)
        if 'status' in data and data['status'] not in self.authorized_status:
            api.abort(500,
                f"Status {data['status']} is not possible (only {' '.join(self.authorized_status)})")
        db.session.add(object)
        db.session.commit()
        return object

    def update(self, id, data):
        object = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(object,attr): 
                if getattr(object,attr)!=value:
                    if attr=='status' and value not in self.authorized_status:
                        api.abort(500,
                           f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                    setattr(object, attr, value)
                    modified = True
            else:
                api.abort(500,f'Error: {object.__name__} has no attribute {attr}')
        if modified:
            if hasattr(object,'modification_date'):
                object.modification_date = datetime.utcnow()
            db.session.commit()
        return object

    def delete(self, id):
        object = self.get(id)
        db.session.delete(object)
        db.session.commit()
        return object
    
    def list(self, sorting_column=None, **args):
        if args:
            final_filter = process_filtering_args(self.ObjectType, args)
            return list(self.ObjectType.query.filter(final_filter).order_by(sorting_column))
        else:
            return list(self.ObjectType.query.order_by(sorting_column).all())
        
        



ns = api.namespace('tasks', description='TASK operations')

class TaskDAO(BaseDAO):
    ObjectType = Task
    authorized_status = ['paused','pending','assigned','accepted','running','failed','succeeded']

    def list(self, **args):
        return super().list(sorting_column='task_id',**args)

    def update(self, id, data):
        task = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(task,attr): 
                if getattr(task,attr)!=value:
                    if attr=='status':
                        if value not in self.authorized_status:
                            api.abort(500,
                                f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                        if task.status=='running':
                            for execution in db.session.query(Execution).filter(
                                    Execution.task_id==id, Execution.status=='running'):
                                db.session.add(Signal(execution.execution_id, execution.worker_id, 9))
                                execution.status='failed'
                                execution.modification_date = datetime.utcnow()
                    setattr(task, attr, value)
                    modified = True
            else:
                raise Exception('Error: {} has no attribute {}'.format(
                    task.__name__, attr))
        if modified:
            task.modification_date = datetime.utcnow()
            db.session.commit()
        return task


task_dao = TaskDAO()

task = api.model('Task', {
    'task_id': fields.Integer(readonly=True, description='The task unique identifier'),
    'name': fields.String(required=False, description='The task (optional) name'),
    'status': fields.String(required=False, 
        description=f'The task status: {", ".join(TaskDAO.authorized_status)}'), 
    'command': fields.String(required=True, description='The task command'), 
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of task creation in server'),
    'modification_date': fields.DateTime(readonly=True,
        description='timestamp of task last modification'),
    'batch': fields.String(required=False, 
        description="only workers with the same batch (null or not) will accept the task."),
    'input': fields.String(required=False,
        decription="Input data required for task (space separated files URL in s3://...)"),
    'output': fields.String(required=False,
        decription="Output data basepath where /output content will be copied (if null, the result transmission is up to the command)"),
    'container': fields.String(required=False,
        decription="Container (as for now this is a docker) in which task is launched"),
    'container_options': fields.String(required=False,
        decription="Container (extra) option if needed"),
    'resource': fields.String(required=False,
        decription="Resource data required for task (much like input except it is shared between tasks) (space separated files URL in s3://...)"),
})

task_filter = api.model('TaskFilter', {
    'task_id': fields.List(fields.Integer(),required=False,decription='A list of ids to restrict listing'),
    'batch': fields.String(required=False, description="Filter with this batch"),
    'status': fields.String(required=False, description="Filter with this status"),
    'name': fields.String(required=False, description="Filter with this name"),
})


@ns.route('/')
class TaskList(Resource):
    '''Shows a list of all tasks, and lets you POST to add new tasks'''
    @ns.doc('list_tasks')
    @ns.expect(task_filter)
    @ns.marshal_list_with(task)
    def get(self):
        '''List all tasks'''
        return task_dao.list(**api.payload) if api.payload else task_dao.list()

    @ns.doc('create_task')
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def post(self):
        '''Create a new task'''
        return task_dao.create(api.payload), 201


@ns.route("/<id>")
@ns.param("id", "The task identifier")
@ns.response(404, "Task not found")
class WorkerObject(Resource):
    @ns.doc("get_task")
    @ns.marshal_with(task)
    def get(self, id):
        """Fetch a task given its identifier"""
        return task_dao.get(id)

    @ns.doc("update_task")
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def put(self, id):
        """Update a task"""
        return task_dao.update(id, api.payload)

    @ns.doc("delete_task")
    @ns.marshal_with(task)
    def delete(self, id):
        """Delete a task"""
        return task_dao.delete(id)

ns = api.namespace('workers', description='WORKER operations')



class WorkerDAO(BaseDAO):
    ObjectType = Worker
    authorized_status = ['paused','running','offline','failed']

    def update_contact(self, id, load,memory,stats):
        db.engine.execute(
            db.update(Worker
                    ).values({'last_contact_date':datetime.utcnow(),
                        'load':load,'memory':memory,'stats':stats}
                    ).where(Worker.worker_id==id)
        )
        db.session.commit()

    def delete(self,id,is_destroyed=False, session=db.session):
        """Delete a worker
        """
        worker=self.get(id)
        log.warning(f'Deleting worker {id} ({worker.idle_callback})')
        if worker.idle_callback is not None and not is_destroyed:
            create_worker_destroy_job(worker, session)
            return worker
        else:
            object = self.get(id)
            session.delete(object)
            session.commit()
            return object

    def list(self):
        return super().list(sorting_column='worker_id')

worker_dao = WorkerDAO()

worker = api.model('Worker', {
    'worker_id': fields.Integer(readonly=True, description='The worker unique identifier'),
    'name': fields.String(required=True, description='The worker name'),
    'concurrency': fields.Integer(required=True, 
        description='The worker concurrency (nb of parallel processes)'), 
    'prefetch': fields.Integer(required=False, 
        description='How many jobs should be prefetch so that they can be launched as soon as possible'),     
    'hostname': fields.String(required=False, description='The worker hostname'),
    'status': fields.String(required=False,
         description=f'The worker status: {", ".join(WorkerDAO.authorized_status)}'), 
    'load': fields.String(readonly=True, description='The worker load (in %)'),
    'memory':fields.Float(readonly=True, description='Memory used (in %)'),
    'stats':fields.String(readonly=True,description='Other worker stats'),
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of worker creation'),
    'modification_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker modification'),
    'last_contact_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker ping (automatically sent by worker'),
    'batch': fields.String(required=False, 
        description="worker accept only tasks with same batch (null or not)."),
    'idle_callback': fields.String(readonly=True,
        description="A command to be called on scitq server when the worker load *returns* to zero. Typically used to end cloud instances.")
})

@ns.route('/')
class WorkerList(Resource):
    '''Shows a list of all workers, and lets you POST to add new workers'''
    @ns.doc('list_workers')
    @ns.marshal_list_with(worker)
    def get(self):
        '''List all workers'''
        return worker_dao.list()

    @ns.doc('create_worker')
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def post(self):
        '''Create a new worker'''
        return worker_dao.create(api.payload), 201


worker_tasks = api.model('WorkerTasks', {
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'status': fields.String(readonly=True, description='The different status of tasks'), 
    'count': fields.Integer(readonly=True, description='How many tasks of this type'),
})

@ns.route("/tasks")
class WorkerTaskList(Resource):
    '''Shows a list of all tasks in all workers, and lets you POST to add new workers'''
    @ns.doc('list_worker_tasks')
    @ns.marshal_list_with(worker_tasks)
    def get(self):
        '''List all workers'''
        return list(db.session.execute("""SELECT w.worker_id, e.status, count(e.task_id) as count 
            FROM worker w
            JOIN execution e ON (e.worker_id=w.worker_id)
            JOIN task t ON (e.task_id=t.task_id 
                AND e.latest)
            GROUP BY w.worker_id,e.status"""))
    

@ns.route("/<id>")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerObject(Resource):
    @ns.doc("get_worker")
    @ns.marshal_with(worker)
    def get(self, id):
        """Fetch a worker given its identifier"""
        return worker_dao.get(id)

    @ns.doc("update_worker")
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def put(self, id):
        """Update a worker"""
        #worker_dao.update_contact(id)
        if 'batch' in api.payload and api.payload['batch']=='':
            api.payload['batch']==None
        return worker_dao.update(id, api.payload)

    @ns.doc("delete_worker")
    @ns.marshal_with(worker)
    def delete(self, id):
        """Delete a worker"""
        return worker_dao.delete(id)

ping_parser = api.parser()
ping_parser.add_argument('load', type=str, help='Worker load', location='json')
ping_parser.add_argument('memory', type=float, help='Worker memory', location='json')
ping_parser.add_argument('stats', type=str, help='Worker other stats', location='json')

@ns.route("/<id>/ping")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerPing(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(ping_parser)
    @ns.marshal_with(worker)
    def put(self, id):
        """Update a worker last contact"""
        args = ping_parser.parse_args()
        worker_dao.update_contact(id, args.get('load',''),args.get('memory',''),args.get('stats',''))
        return worker_dao.get(id)

callback_parser = api.parser()
callback_parser.add_argument('message', type=str, help='Callback message sent (idle)', location='json')

@ns.route("/<id>/callback")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerCallback(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(callback_parser)
    def put(self, id):
        """Update a worker last contact"""
        message = callback_parser.parse_args().get('message','')
        worker = worker_dao.get(id)
        if message == 'idle' and worker.idle_callback:
            if db.session.query(Execution).filter(Execution.status=='running',
                    Execution.worker_id==worker.worker_id).count()>0:
                log.warning(f'Worker {worker.name} called idle callback but some tasks are still running, refusing...')
                return {'result':'still have running tasks'}
            if db.session.query(Task).filter(and_(Task.status.in_(['pending']),
                                    Task.batch==worker.batch)).count()>0:
                log.warning(f'Worker {worker.name} called idle but some tasks are still due...')
                return {'result':'still some work to do, lazy one!'}
            log.warning(f'Worker {worker.name} ({worker.worker_id}) called idle callback, launching: '+worker.idle_callback.format(**(worker.__dict__)))
            #worker.destroy()
            create_worker_destroy_job(worker, db.session, commit=False)
            #db.session.delete(worker)
            db.session.commit()
            return {'result':'ok'}
        else:
            log.warning(f'Worker {worker.name} called idle callback but has no idle command')
            return {'result':'nothing to do'}

deploy_parser = api.parser()
deploy_parser.add_argument('number', type=int, help='How many workers should be deployed', location='json')
deploy_parser.add_argument('region', type=str, help='Which provider region for worker', location='json')
deploy_parser.add_argument('provider', type=str, help='Specify the worker provider', location='json')
deploy_parser.add_argument('flavor', type=str, help='Which provider flavor/model for worker', location='json')
deploy_parser.add_argument('batch', type=str, help='Batch name (that must be shared by tasks) for worker', location='json')
deploy_parser.add_argument('concurrency', type=int, help='How many tasks should be run in parallel', location='json')
deploy_parser.add_argument('prefetch', type=int, help='How many extra tasks should be prefetched', location='json')

@ns.route("/deploy")
class WorkerDeploy(Resource):
    @ns.doc("deploy_worker_vm_and_process")
    @ns.expect(deploy_parser)
    def put(self):
        """Create and deploy one or several workers"""
        deploy_args = deploy_parser.parse_args()

        for _ in range(deploy_args['number']):
            db.session.add(
                Job(target='', 
                    action='worker_create', 
                    args={
                        'concurrency': deploy_args['concurrency'], 
                        'prefetch':deploy_args['prefetch'],
                        'flavor':deploy_args['flavor'],
                        'region':deploy_args['region'],
                        'provider':deploy_args['provider'],
                        'batch':deploy_args['batch']
                    }
                )
            )
        db.session.commit()


        return {'result':'ok'}

class ExecutionDAO(BaseDAO):
    ObjectType = Execution
    authorized_status = ['pending','accepted','running','failed','succeeded']

    def create(self, data):
        task = task_dao.get(data['task_id'])
        if task.status not in ['pending','paused']:
            api.abort(500, f"A new execution for task {task.task_id} is not possible")
        if 'status' in data:
            if data['status']=='running':
                task.status = 'running'
            elif data['status'] in ['pending','paused']:
                task.status = 'assigned'
            else:
                api.abort(500, f"A new execution cannot be created with a status {data['status']}")
        else:
            task.status='assigned'
        return super().create(data)

    def update(self, id, data):
        execution = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(execution,attr): 
                if getattr(execution,attr)!=value:
                    if attr=='status':
                        if value not in self.authorized_status:
                            api.abort(500,
                                f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                        task=task_dao.get(execution.task_id)
                        if execution.status=='pending':
                            if value=='running':
                                task.status = 'running'
                                task.modification_date = datetime.utcnow()
                            elif value=='accepted':
                                task.status = 'accepted'
                                task.modification_date = datetime.utcnow()
                            elif value in ['refused','failed']:
                                task.status = 'pending'
                                task.modification_date = datetime.utcnow()
                            elif value=='succeeded':
                                task.status = 'succeeded'
                                task.modification_date = datetime.utcnow()
                            else:
                                api.abort(500, f"An execution cannot change status from pending to {value}")
                        elif execution.status=='accepted':
                            if value in ['running','failed','succeeded','pending']:
                                task.status = value
                                task.modification_date = datetime.utcnow()
                            else:
                                log.exception(f"An execution cannot change status from accepted to {value}")
                                api.abort(500, f"An execution cannot change status from accepted to {value}")
                        elif execution.status=='running':
                            if value in ['succeeded', 'failed']:
                                task.status=value
                                task.modification_date = datetime.utcnow()
                            else:
                                log.exception(f"An execution cannot change status from running to {value}")
                                api.abort(500, f"An execution cannot change status from running to {value}")
                        else:
                            api.abort(500, f"An execution cannot change status from {execution.status} (only from pending, running or accepted)")
                    setattr(execution, attr, value)
                    modified = True
            else:
                raise Exception('Error: {} has no attribute {}'.format(
                    execution.__name__, attr))
        if modified:
            execution.modification_date = datetime.utcnow()
            db.session.commit()
        return execution

    def list(self,no_output=False,**args):
        sorting_column='execution_id'
        q=Execution.query
        if args:
            q=q.filter(process_filtering_args(Execution, args))
        if no_output:
            q=q.with_entities(Execution.execution_id,
                              Execution.command, 
                              Execution.status,
                              Execution.task_id,
                              Execution.creation_date,
                              Execution.modification_date,
                              Execution.pid,
                              Execution.output_files,
                              Execution.latest)

        return list(q.order_by(sorting_column).all())
 
execution_dao = ExecutionDAO()

execution = api.model('Execution', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(required=True, description='A worker unique identifier'),
    'task_id': fields.Integer(required=True, description='A task unique identifier'),
    'status': fields.String(readonly=True,
        description=f'The execution status: {", ".join(ExecutionDAO.authorized_status)}'), 
    'return_code': fields.Integer(required=False, description='The return code of the execution (when finished)'),
    'pid': fields.String(required=False, description='The process id (pid) of the execution'),
    'creation_date': fields.DateTime(readonly=True,
        description="timestamp of execution creation (on worker)"),
    'modification_date': fields.DateTime(readonly=True,
        description="timestamp of execution last modification"),
    'output': fields.String(readonly=True, description='The standard output of the execution'),
    'error': fields.String(readonly=True, description='The standard error of the execution (if any)'),
    'output_files': fields.String(readonly=True, description='A list of output files transmitted (if any)'),
    'command': fields.String(required=False, description='The command that was really launched for this execution (it case Task.execution is modified)'),
    'latest': fields.Boolean(readonly=True, description='Latest or current execution for the related task')
})


@ns.route("/<id>/executions")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerExecutionObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution)
    def get(self, id):
        """Fetch a worker executions given the worker identifier"""
        #worker_dao.update_contact(id)
        return execution_dao.list(worker_id=id, no_output=True)


@ns.route("/<id>/executions/<status>")
@ns.param("id", "The worker identifier")
@ns.param("status", "Filter only executions with this status")
@ns.response(404, "Worker not found")
class WorkerExecutionFilterObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution)
    def get(self, id, status):
        """Fetch a worker executions given the worker identifier and the executions status"""
        #worker_dao.update_contact(id)
        return execution_dao.list(worker_id=id, status=status)

signal = api.model('Signal', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'signal': fields.Integer(readonly=True, description='The signal to send to the execution (UNIX signal)'),
})

signal_parser = api.parser()
signal_parser.add_argument('execution_id', type=int, help='Execution identifier', location='json')
signal_parser.add_argument('signal', type=int, help='Signal to send', location='json')

@ns.route("/<id>/signals")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerSignal(Resource):
    @ns.doc("get_worker_signals")
    @ns.marshal_list_with(signal)
    def get(self, id):
        """Fetch a worker executions given the worker identifier and the executions status"""
        signals = list(Signal.query.filter(Signal.worker_id==id))
        for sig in signals:
            db.session.delete(sig)
        db.session.commit()
        return signals
    
    @ns.doc("create_worker_signal")
    @ns.expect(signal_parser)
    def post(self, id):
        """Create a signal for the given worker about a certain execution with a certain signal"""
        args = signal_parser.parse_args()
        db.session.add(Signal(args['execution_id'], id, args['signal']))
        db.session.commit()


ns = api.namespace('executions', description='EXECUTION operations')

execution_filter = api.model('ExecutionFilter', {
    'task_id': fields.Integer(required=False,decription='A list of ids to restrict listing'),
    'status': fields.String(required=False, description="Filter with this status"),
    'latest': fields.Boolean(required=False, decription="Filter only for latest execution")
})
@ns.route('/')
class ExecutionList(Resource):
    '''Shows a list of all executions, and lets you POST to add new workers'''
    @ns.doc('list_executions')
    @ns.expect(execution_filter)
    @ns.marshal_list_with(execution)
    def get(self):
        '''List all executions'''
        return execution_dao.list(**api.payload) if api.payload else execution_dao.list()

    @ns.doc('create_execution')
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def post(self):
        '''Create a new execution'''
        return execution_dao.create(api.payload), 201
    

@ns.route("/<id>")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc("get_execution")
    @ns.marshal_with(execution)
    def get(self, id):
        """Fetch a execution given its identifier"""
        return execution_dao.get(id)

    @ns.doc("update_execution")
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def put(self, id):
        """Update an execution"""
        return execution_dao.update(id, api.payload)


parser = api.parser()
parser.add_argument('text', type=str, help='Supplementary text', location='json')

@ns.route("/<id>/delete")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc('delete_execution')
    def put(self,id):
        '''Delete an execution for this id'''
        for e in Execution.query.filter(Execution.execution_id==id):
            db.session.delete(e)
        db.session.commit()
        return {'result':'ok'}

@ns.route("/<id>/output")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to the execution output"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output':('' if x.output is None else x.output) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/error")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_error")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to execution error"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'error':('' if x.error is None else x.error) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/output_files")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutputFiles(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Update output files (files transferred from /output, i.e. results) for execution as a space separated file URL list"""
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output_files':args['text']})
        return {'result':'Ok'}


ns = api.namespace('batch', description='BATCH level operations')

batch_parser = api.parser()
batch_parser.add_argument('signal', type=int, help='Optionnaly send signal to all running tasks', 
    required=False, location='json')


custom_batch = api.model('Batch', {
    'batch': fields.String(readonly=True, description='The name of the batch'), 
    'pending': fields.Integer(readonly=True,
        description='The number of task in pending state'), 
    'accepted': fields.Integer(readonly=True,
        description='The number of task in accepted state'), 
    'running': fields.Integer(readonly=True,
        description='The number of task in running state'), 
    'failed': fields.Integer(readonly=True,
        description='The number of task in succeeded state'), 
    'succeeded': fields.Integer(readonly=True,
        description='The number of task in failed state'), 
    'workers': fields.String(readonly=True, description='The workers used in that batch'),
})

@ns.route("/")
class BatchList(Resource):
    @ns.doc("list_batch")
    @ns.marshal_list_with(custom_batch)
    def get(self):
        """List the batches, their task statuses and workers"""
        if isinstance(db.session.bind.dialect, sqlite.dialect):
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') FROM worker GROUP BY batch'''
        else:
            worker_query='''SELECT batch,STRING_AGG(name,',') FROM worker GROUP BY batch'''
        batch_query='''SELECT batch,status,count(task_id) FROM task GROUP BY batch,status ORDER BY batch,status'''
        batches = []
        batches_attributes = {}
        for batch,status,count in db.session.execute(batch_query):
            if batch not in batches:
                batches.append(batch)
                batches_attributes[batch]={'batch':batch}    
            batches_attributes[batch][status]=count
        for batch,workers in db.session.execute(worker_query):
            if batch in batches_attributes:
                batches_attributes[batch]['workers']=workers
        return [batches_attributes[batch] for batch in batches]
@ns.route("/<name>/stop")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchStop(Resource):
    @ns.doc("stop_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """Pause all workers for this batch - and kill current job if force is set"""
        if name=='Default':
            for w in Worker.query.filter(Worker.batch.is_(None)):
                w.status = 'paused'
        else:
            for w in Worker.query.filter(Worker.batch==name):
                w.status = 'paused'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            if name=='Default':
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch.is_(None))):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
            else:
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch==name)):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}

@ns.route("/<name>/go")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchGo(Resource):
    @ns.doc("_re_launch_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """(re)set all workers affected to this batch to running"""
        if name=='Default':
            for w in Worker.query.filter(Worker.batch.is_(None)):
                w.status = 'running'
        else:
            for w in Worker.query.filter(Worker.batch==name):
                w.status = 'running'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            if name=='Default':
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch.is_(None))):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
            else:
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch==name)):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}

@ns.route("/<name>")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchDelete(Resource):
    @ns.doc("delete_a_batch")
    def delete(self, name):
        """Delete all tasks and executions for this batch"""
        if name=='Default':
            log.warning('Deleting default batch')
            db.session.execute(delete(Execution).where(Execution.task_id.in_(
                select(Task.task_id).where(Task.batch.is_(None)))), 
                execution_options={'synchronize_session':False})
            db.session.execute(delete(Task).where(Task.batch.is_(None)))
        else:
            log.warning(f'Deleting batch {name}')
            db.session.execute(delete(Execution).where(Execution.task_id.in_(
                select(Task.task_id).where(Task.batch==name))), 
                execution_options={'synchronize_session':False})
            db.session.execute(delete(Task).where(Task.batch==name))
        db.session.commit()
        return {'result':'Ok'}



#     # ### 
#     #  #  
#     #  #  
#     #  #  
#     #  #  
#     #  #  
 #####  ### 


package_version = package_version()
@app.route('/ui/')
def ui(name=None):
    return render_template('ui.html', name=name, package_version=package_version)
@app.route('/ui/task/')
def task():
    return render_template('task.html',package_version=package_version)

@app.route('/ui/batch/')
def batch():
    return render_template('batch.html', package_version=package_version)



@app.route('/ui/get/')
def handle_get():
    json = request.args
    if 'delay' in json:
        sleep(int(json['delay']))
    if json['object']=='workers':
        log.info('sending workers')
        return jsonify({
            'workers':list([tryupdate(dict(row),'stats',json_module.loads,row.stats) for row in db.session.execute(
                '''SELECT 
                    worker_id,
                    name, 
                    batch,
                    status, 
                    concurrency,
                    prefetch,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND
                                                                        execution.status='accepted') as accepted,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='running') as running,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='succeeded') as succeeded,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='failed') as failed,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id) as total,
                    load,
                    memory,
                    stats
                FROM worker
                ORDER BY worker.batch,worker.name''')]),
            'totals': next(iter([dict(row) for row in db.session.execute(
                '''SELECT
                    (SELECT count(task_id) FROM task WHERE status='pending') as pending,
                    (SELECT count(task_id) FROM task WHERE status IN ('assigned','accepted')) as assigned,
                    (SELECT count(task_id) FROM task WHERE status='running') as running,
                    (SELECT count(task_id) FROM task WHERE status='failed') as failed,
                    (SELECT count(task_id) FROM task WHERE status='succeeded') as succeeded
                '''
            )])) })

    elif json['object']=='tasks':
        order_by = json.get('order_by', None)
        filter_by = json.get('filter_by', None)
        worker = json.get('worker', None)

        log.warning(f"sending task ordered by {order_by} filtered by {filter_by} for worker {worker}")


        if order_by=='worker':
            sort_clause='ORDER BY worker.name, task.task_id DESC'
        elif order_by=='batch':
            sort_clause='ORDER BY task.batch, task.task_id DESC'
        else:
            sort_clause='ORDER BY task.task_id DESC'
        
        where_clauses = []

        if filter_by=='terminated':
            where_clauses.append("execution.status IN ('succeeded','failed')")
        elif filter_by:
            where_clauses.append(f"execution.status='{filter_by}'")

        if worker is not None:
            try:
                worker=int(worker)
                where_clauses.append(f"execution.worker_id={worker}")
            except:
                log.error(f'task filtering with worker is not possible: {worker} is not a valid id')

        where_clause = f'''WHERE {' AND '.join(where_clauses)}''' if where_clauses else ''

        if IS_SQLITE:
            trunc_output=f'SUBSTR(execution.output,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
            trunc_error=f'SUBSTR(execution.error,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
        else:
            trunc_output=f'RIGHT(execution.output,{UI_OUTPUT_TRUNC})'
            trunc_error=f'RIGHT(execution.error,{UI_OUTPUT_TRUNC})'
        

        #task_list = list([list(map(lambda x : str(x) if type(x)==type(datetime.utcnow()) else x ,row)) for row in db.session.execute(
        task_list = list([dict(row) for row in db.session.execute(
        f'''SELECT
        task.task_id,
        task.name,
        worker.name as worker_name,
        task.batch,
        execution.creation_date,
        execution.modification_date,
        execution.execution_id,
        {trunc_output} as output,
        {trunc_error} as error,
        task.command,
        execution.worker_id,
        task.status
        FROM task 
        LEFT JOIN execution ON (task.task_id=execution.task_id AND latest)
        LEFT JOIN worker ON execution.worker_id=worker.worker_id 
        {where_clause}
        {sort_clause}
        LIMIT {UI_MAX_DISPLAYED_ROW}
        '''
        )])
        
        detailed_tasks = json.getlist('detailed_tasks[]')
        if detailed_tasks:
            log.warning(f"detailing tasks {detailed_tasks}")
            for detailed_task in db.session.execute(f"""
                SELECT execution_id,output,error FROM execution 
                WHERE execution_id IN ({','.join([str(eid) for eid in detailed_tasks])})"""):
                for task in task_list:
                    if task['execution_id']==detailed_task['execution_id']:
                        task['output']=detailed_task['output']
                        task['error']=detailed_task['error']
                        break

        return jsonify({'tasks':task_list})
        
    elif json['object'] == 'batch':
        log.info('sending batch')
        if IS_SQLITE:
            duration_query='(JULIANDAY(e1.modification_date)-JULIANDAY(e1.creation_date))*24'
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') as workers FROM worker GROUP BY batch ORDER BY name'''
        else:
            duration_query='EXTRACT ( EPOCH FROM (e1.modification_date-e1.creation_date)/3600 )'
            worker_query='''SELECT batch,STRING_AGG(name,',') as workers FROM (SELECT * FROM worker ORDER BY name) w GROUP BY batch'''
        batch_query=f'''SELECT * FROM (
    SELECT batch,status,COUNT(task_id) as count,MAX(duration) as max,MIN(duration) as min, AVG(duration) as avg FROM (
        SELECT {duration_query} as duration, e1.task_id, e1.status,task.batch FROM execution e1 JOIN task ON (
            task.task_id=e1.task_id AND e1.latest
        )
    ) AS e2 GROUP BY batch,status
    UNION	 
    SELECT batch,status, COUNT(task_id),NULL,NULL,NULL 
    FROM task WHERE task_id NOT IN (SELECT task_id FROM execution) GROUP BY batch,status
) AS b ORDER BY batch, status'''
        return jsonify({'batches':list([dict(row) for row in db.session.execute(batch_query)]),
                        'workers': list([dict(row) for row in db.session.execute(worker_query)])})

#@socketio.on('change_batch')
@app.route('/ui/change_batch')
def handle_change_batch():
    json = request.args
    Worker.query.filter(Worker.worker_id==json['worker_id']).update(
        {Worker.batch:json['batch_name'] or None})
    db.session.commit()
    return '"ok"'


#@socketio.on('concurrency_change')
@app.route('/ui/concurrency_change')
def handle_concurrency_change():
    json = request.args
    worker_id = json['id']
    change = json['change']
    log.info(f'changing concurrency for worker {worker_id}: {change}')
    if isinstance(db.session.bind.dialect, sqlite.dialect):
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.max(Worker.concurrency+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.greatest(Worker.concurrency+change,0)})
    db.session.commit()
    return '"ok"'

#@socketio.on('prefetch_change')
@app.route('/ui/prefetch_change')
def handle_prefetch_change():
    json = request.args
    worker_id = json['id']
    change = json['change']
    log.info(f'changing prefetch for worker {worker_id}: {change}')
    if isinstance(db.session.bind.dialect, sqlite.dialect):
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.max(Worker.prefetch+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.greatest(Worker.prefetch+change,0)})
    db.session.commit()
    return '"ok"'

@app.route('/ui/pause_unpause_worker')
def handle_pause_unpause_worker():
    json = request.args
    worker_id = json['id']
    status = json['status']
    Worker.query.filter(Worker.worker_id==worker_id).update({Worker.status:status})
    log.info(f'changing status for worker {worker_id}: {status}')
    db.session.commit()
    return '"ok"'


@app.route('/ui/clean_worker')
def handle_clean_worker():
    json = request.args
    worker_id = json['worker_id']
    db.session.add(Signal(execution_id=None, worker_id=worker_id, signal=SIGNAL_CLEAN))
    db.session.commit()
    log.info(f'sending clean signal for worker {worker_id}')
    return '"ok"'

@app.route('/ui/restart_worker')
def handle_restart_worker():
    json = request.args
    worker_id = json['worker_id']
    db.session.add(Signal(execution_id=None, worker_id=worker_id, signal=SIGNAL_RESTART))
    db.session.commit()
    log.info(f'sending restart signal for worker {worker_id}')
    return '"ok"'

#@socketio.on('create_worker')
@app.route('/ui/create_worker')
def handle_create_worker():
    json = request.args
    concurrency = int(json['concurrency'])
    flavor = json['flavor']
    if not flavor:
        return jsonify(error='Flavor must be specified')
        #return None
    region = json['region']
    provider = json['provider']
    if not region:
        return jsonify(error='Region must be specified')
        #return None
    batch = json['batch'] or None
    prefetch = int(json['prefetch'])
    number = int(json['number'])
    for _ in range(number):
        db.session.add(
            Job(target='', 
                action='worker_create', 
                args={
                    'concurrency': concurrency, 
                    'prefetch':prefetch,
                    'flavor':flavor,
                    'region':region,
                    'provider': provider,
                    'batch':batch
                }
            )
        )
    db.session.commit()
    return '"ok"'

#@socketio.on('batch_action')
@app.route('/ui/batch/action')
def handle_batch_action():
    """Gathering all the action dealing with batch like pause, break, stop, clear, go."""
    json = request.args
    if json['action'] in ['stop','break','pause','simple pause','pause only batch']:
        #Same function as in the API set all workers affected to this batch to running and can also interrupt the running tasks with signal 3 and 9
        name=json['name']
        if name=='Default':
            name=None
        for w in Worker.query.filter(Worker.batch==name):
                w.status = 'paused'
        if json['action']=='break':
            signal = 9
        elif json['action']=='stop':
            signal = 3
        elif json['action']=='simple pause':
            signal = 0
        elif json['action']=='pause':
            #TODO: make a more efficient query
            for t in Task.query.filter(and_(
                        Task.batch==name,
                        Task.status.in_(['running','accepted']))):
                t.status = 'paused'
            signal = 20
            db.session.commit()
        log.warning(f'Sending signal {signal} to executions for batch {name}')
        for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                        Execution.status=='running',
                                        Task.batch==name)):
            log.warning(f'Sending signal {signal} to execution {e.execution_id}')
            db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result pause :Ok')
    elif json['action'] in ['go','simple go']: 
        #Same function as in the API (re)set all workers affected to this batch to running
        name=json['name']
        """(re)set all workers affected to this batch to running"""
        for w in Worker.query.filter(Worker.batch==name):
            w.status = 'running'
        if json['action']=='go':
            signal=18
            for t in Task.query.filter(Task.batch==name):
                if t.status == 'paused':
                    t.status='running'
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.batch==name)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result go : Ok')
    elif json['action']=='clear':
        #Same function as in the API clear() Delete all tasks and executions for this batch
        name=json['name']
        """Delete all tasks and executions for this batch"""
        # execution are deleted by cascade
        if name=='Default':
            name=None
        for t in Task.query.filter(Task.batch==name):
            db.session.delete(t)
        db.session.commit()
        log.warning(f'result clear batch {name}: Ok ')
    return '"ok"'

#@socketio.on('task_action')
@app.route('/ui/task/action')
def handle_task_action():
    """Gathering all the action dealing with task like break, stop, delete, modify, restart"""
    #The code essentially is from the API code with a few modifications
    json = request.args
    task=json['task_id']
    if json['action'] in ['break','stop','pause','resume']: 
        #A signal 3 or 9 is created and causes only the interruption of the task with id same structure as in the API 
        task=json['task_id']
        for t in Task.query.filter(Task.task_id==task):
            if json['action'] == 'break':
                type='break'
                signal = 9
            elif json['action'] == 'stop':
                signal = 3
                type='stop'
            elif json['action'] == 'pause':
                signal = 20
                type='pause'
                t.status='paused'
            elif json['action'] == 'resume':
                signal = 18
                type='resume'
                t.status='running'
            log.warning(f'Sending signal {signal} to executions for task {task}')
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.task_id==task)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning(f'result {type} : Ok')
    elif json['action']=='delete': 
        #Delete the task in the data base
        for t in Task.query.filter(Task.task_id==task):
            db.session.delete(t)
        db.session.commit()
        log.warning('result delete: Ok')
    elif json['action']=='modify': 
        #Changing the command for a task in the data base and moving it in the task queue. It doesn't create a new task.
        for t in Task.query.filter(Task.task_id==task):
            t.command =json["modification"]
            t.status='pending'
        db.session.commit()
        log.warning('result modify : Ok')
    elif json['action']=='restart': 
        #Relaunching the execution of a task.
        for t in Task.query.filter(Task.task_id==task):
            t.status='pending'
        db.session.commit()
        log.warning('result restart : Ok')
    return '"ok"'

#@socketio.on('delete_worker') #Delete a worker.
@app.route('/ui/delete_worker')
def delete_worker():
    """Delete a worker in db"""
    json = request.args
    worker_dao.delete(json['worker_id'], session=db.session)
    return '"ok"'
    

#@socketio.on('jobs')
@app.route('/ui/jobs')
def handle_jobs():
    """Provide UI with job list"""
    return jsonify(jobs = [to_dict(job) for job in db.session.query(Job).order_by(Job.job_id.desc()).all()])

@app.route('/ui/delete_job')
def delete_job():
    """Delete a job in db"""
    json = request.args
    job = db.session.query(Job).get(json['job_id'])
    if job:
        db.session.delete(job)
        db.session.commit()
    else:
        log.warning(f"Job {json['job_id']} already deleted")
    return '"ok"'

@app.route('/ui/delete_jobs')
def delete_jobs():
    """Delete terminated jobs"""
    if db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='succeeded')
            ).scalar_one()>0:
        log.warning('Here')
        db.session.execute(delete(Job).where(Job.status=='succeeded'))
    elif db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='failed')
            ).scalar_one()>0:
        log.warning('There')
        db.session.execute(delete(Job).where(Job.status=='failed'))
    elif db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='pending')
            ).scalar_one()>0:
        db.session.execute(delete(Job).where(Job.status=='pending'))
    else:
        return '"nothing to do"'
    db.session.commit()
    return '"ok"'

@app.route('/ui/restart_job')
def restart_job():
    """Restart a job in db"""
    json = request.args
    job = db.session.query(Job).get(json['job_id'])
    if job:
        job.status = 'pending'
        job.retry = 0
        db.session.commit()
    else:
        log.warning(f"Job {json['job_id']} already deleted")
    return '"ok"'

#ns = api.namespace('live', description='Very basic API for the UI')

@app.route('/live/ping')
def live_ping():
    """Simply return ok"""
    return '"ok"'

 #####     ##     ####   #    #   ####   #####    ####   #    #  #####       #####  #    #  #####   ######    ##    #####  
 #    #   #  #   #    #  #   #   #    #  #    #  #    #  #    #  #    #        #    #    #  #    #  #        #  #   #    # 
 #####   #    #  #       ####    #       #    #  #    #  #    #  #    #        #    ######  #    #  #####   #    #  #    # 
 #    #  ######  #       #  #    #  ###  #####   #    #  #    #  #    #        #    #    #  #####   #       ######  #    # 
 #    #  #    #  #    #  #   #   #    #  #   #   #    #  #    #  #    #        #    #    #  #   #   #       #    #  #    # 
 #####   #    #   ####   #    #   ####   #    #   ####    ####   #####         #    #    #  #    #  ######  #    #  #####  


def get_nodename(session):
    worker_names = list(map(lambda x: x[0], 
        session.execute(select(Worker.name))))
    log.warning(f'Worker names: {worker_names}')
    i=1
    while f'node{i}' in worker_names:
        i+=1
    return f'node{i}'


def create_worker_object(concurrency, flavor, region, provider, batch, prefetch, db_session):
    """Create a worker object in db - this must be called linearly not in async way
    """
    hostname = get_nodename(db_session)
    idle_callback = WORKER_IDLE_CALLBACK.format(hostname=hostname)
    log.info(f'Creating a new worker {hostname}: concurrency:{concurrency}, \
flavor:{flavor}, region:{region}, provider:{provider}, prefetch:{prefetch}')
    w = Worker(name=hostname, hostname=hostname, concurrency=concurrency, status='offline', 
            batch=batch, idle_callback=idle_callback, prefetch=prefetch)
    db_session.add(w)
    db_session.commit()
    return w
    

def background():
    # while some tasks are pending without executions:
    #   look for a running worker:
    #      create a pending execution of this task for this worker


    with app.app_context():
        session = Session(db.engine)
    worker_process_queue = {}
    other_process_queue = []
    log.info('Starting thread for {}'.format(os.getpid()))
    ansible_workers = list(session.query(Worker).filter(and_(
                    Worker.status=='running',
                    Worker.idle_callback.is_not(None))).with_entities(
                        Worker.hostname))
    if ansible_workers:
        log.warning(f'Making sure workers {",".join([w.hostname for w in ansible_workers])} has access to server')
        process = PropagatingThread(
                    target=run,
                    args = (SERVER_CRASH_WORKER_RECOVERY,),
                    kwargs= {'shell': True, 'check':True}
                )
        other_process_queue.append(('Worker access task',process))
        process.start()

    while True:
        log.warning('Starting main loop')
        try:
            task_list = list(session.query(Task).filter(
                    Task.status=='pending').with_entities(Task.task_id, Task.batch))
            if task_list:
                task_attributions = False
                worker_list = list(session.query(Worker).filter(
                            Worker.status=='running').with_entities(
                            Worker.worker_id,Worker.batch,Worker.concurrency,Worker.prefetch))
                execution_per_worker = {worker_id: count for worker_id,count in 
                                session.query(Execution.worker_id,func.count(Execution.task_id)).filter(and_(
                                    Execution.worker_id.in_(list([w.worker_id for w in worker_list])),
                                    Execution.status.in_(['running','pending','accepted']))
                                ).group_by(
                                    Execution.worker_id
                                )
                }
                
                for task in task_list:
                    for worker in worker_list:
                        if worker.batch != task.batch:
                            continue
                        if execution_per_worker.get(worker.worker_id,0)<(worker.concurrency+worker.prefetch):
                            session.add(Execution(worker_id=worker.worker_id,
                                task_id=task.task_id))
                            session.query(Task).filter(Task.task_id==task.task_id).update(
                                {'status':'assigned'}
                            )
                            execution_per_worker[worker.worker_id] = execution_per_worker.get(worker.worker_id,0)+1
                            log.info(f'Execution of task {task.task_id} proposed to worker {worker.worker_id}')
                            task_attributions = True
                            break
                if task_attributions:
                    session.commit()
            now = datetime.utcnow()
            log.warning('Looking for offline/online workers')
            session.expire_all()
            for worker in session.query(Worker).filter(Worker.status.in_(['offline','running'])):
                change = False
                log.warning(f'Assessing worker {worker.name}')
                if worker.last_contact_date is None or (
                            now - worker.last_contact_date
                        ).total_seconds() > WORKER_OFFLINE_DELAY:
                    log.warning(f'... late: {worker.last_contact_date}')
                    if worker.status=='running':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) lost, marked as offline')
                        worker.status = 'offline'
                        change = True
                else:
                    log.warning('... good')
                    if worker.status=='offline':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) recovered, marked as running')
                        worker.status = 'running'
                        change = True
                if change:
                    session.commit()
            
            change = False
            for job in list(session.query(Job).filter(Job.status == 'pending')):

                if job.action == 'worker_destroy':
                    change=True
                    if ('create',job.target) in worker_process_queue:
                        worker,worker_create_process,job_id = worker_process_queue[('create',job.target)]
                        if worker_create_process.poll() is None:
                            worker_create_process.terminate()
                            log.warning(f'Worker {job.target} creation process has been terminated')
                            del(worker_process_queue[('create',job.target)])
                            session.query(Job).get(job_id).status='failed'
                    worker = Namespace(**job.args)
                    host_exist_in_ansible = bool(json_module.loads(scitq_inventory(host=job.target)))
                    if host_exist_in_ansible:
                        if len(worker_process_queue)<WORKER_CREATE_CONCURRENCY:
                            log.warning(f'Launching destroy process for {job.target}, command is "{worker.idle_callback}"')
                            worker_delete_process = Popen(
                                    worker.idle_callback,
                                    stdout = PIPE,
                                    stderr = PIPE,
                                    shell = True,
                                    encoding = 'utf-8'
                                )
                            job.status='running'
                            worker_process_queue[('destroy',job.target)]=(worker, worker_delete_process, job.job_id)
                            log.warning(f'Worker {job.target} destruction process has been launched')
                    else:
                        log.warning(f'Deleting worker {worker.name} ({worker.worker_id})')
                        real_worker = session.query(Worker).get(worker.worker_id)
                        if real_worker is not None:
                            change = True
                            session.delete(real_worker)
                            job.status='succeeded'
                
                if job.action == 'worker_create':
                    change = True
                    worker = create_worker_object(db_session=session,
                        **job.args)
                    job.action = 'worker_deploy'
                    job.target = worker.name
                    job.args = dict(job.args)
                    job.args['worker_id'] = worker.worker_id
                    job.retry = WORKER_CREATE_RETRY
                    job.status = 'pending'
                
                if job.action == 'worker_deploy':
                    if job.target not in worker_process_queue and len(
                                worker_process_queue)<WORKER_CREATE_CONCURRENCY:
                        if ('destroy',job.target) in worker_process_queue:
                            log.warning(f'Trying to recreate worker {job.target} after destruction too soon, waiting a little bit...')
                            continue
                        change = True
                        log.warning(f'Launching creation process for worker {job.target}.')
                        worker = Namespace(**job.args)
                        log.warning(f'Launching command is "'+WORKER_CREATE.format(
                                hostname=job.target,
                                concurrency=worker.concurrency,
                                flavor=worker.flavor,
                                region=worker.region,
                                provider=worker.provider,
                            )+'"')
                        worker_create_process = Popen(
                            WORKER_CREATE.format(
                                hostname=job.target,
                                concurrency=worker.concurrency,
                                flavor=worker.flavor,
                                region=worker.region,
                                provider=worker.provider
                            ),
                            stdout = PIPE,
                            stderr = PIPE,
                            shell = True,
                            encoding = 'utf-8'
                        )
                        worker.name = worker.hostname = job.target
                        job.status = 'running'
                        worker_process_queue[('create',job.target)]=(worker, worker_create_process, job.job_id)
                        log.warning(f'Worker {job.target} creation process has been launched')

            if change:
                session.commit()                

            change = False
            for ((action,worker_name),(worker,worker_process,job_id)) in list(worker_process_queue.items()):
                returncode = worker_process.poll()
                if returncode is not None:
                    change = True
                    job = session.query(Job).get(job_id)
                    del(worker_process_queue[(action,worker_name)])
                    if returncode == 0:
                        log.warning(f'Process {action} succeeded for worker {worker.name}.')
                        job.log = worker_process.stdout.read()
                        job.status = 'succeeded'

                        if action=='destroy':
                            #session.execute(Worker.__table__.delete().where(
                            #    Worker.__table__.c.worker_id==worker.worker_id))
                            log.warning(f'Deleting worker {worker.name} ({worker.worker_id})')
                            real_worker = session.query(Worker).get(worker.worker_id)
                            if real_worker is not None:
                                session.delete(real_worker)
                            else:
                                log.error(f'Could not find a worker with worker_id {worker.worker_id}')
                            #worker_dao.delete(worker.worker_id, is_destroyed=True)
                    else:
                        stderr = worker_process.stderr.read()
                        log.warning(f'Process {action} failed for worker {worker.name}: {stderr}')
                        job.log = worker_process.stdout.read() + stderr
                        log.warning(f'Job output is {job.log}')
                        if job.retry > 0:
                            job.retry -= 1
                            job.status = 'pending'
                        else:
                            job.status = 'failed'
                            if action=='create':
                                worker = session.query(Worker).get(job.args['worker_id'])
                                worker.status = 'failed'

            for job in list(session.query(Job).filter(Job.status == 'running')):
                if job.action=='worker_deploy':
                    action='create'
                elif job.action=='worker_destroy':
                    action='destroy'
                else:
                    action=job.action
                if (action, job.target) not in worker_process_queue:
                    log.warning(f'Job {(job.action, job.target)} seems to have failed, not in {worker_process_queue}')
                    job.status='failed'
                    change = True

            if change:
                session.commit()
                
            for process_name, process in list(other_process_queue):
                try:
                    if not process.is_alive():
                        other_process_queue.remove((process_name, process))
                        log.warning(f'Job {process_name} is done.')
                except Exception as e:
                    log.exception(f'Job {process_name} failed: {e}')
                    other_process_queue.remove((process_name, process))

                    

            



                
        except Exception:
            log.exception('An exception occured during server main loop:')
            while True:
                sleep(MAIN_THREAD_SLEEP)
                try:
                    log.warning('Trying to reconnect...')
                    with app.app_context():
                        session = Session(db.engine)
                    break
                except Exception:
                    pass
        sleep(MAIN_THREAD_SLEEP)

if not os.environ.get('SCITQ_PRODUCTION'):
    Thread(target=background).start()


def main():
    app.run(debug=True)

if __name__ == "__main__":
    #raise RuntimeError('Do not launch directly, launch with "FLASK=scitq.server flask run"')
    #main()
    pass


import re
import requests
from requests.exceptions import Timeout, ConnectionError
from datetime import datetime
from time import sleep
import threading
import queue
from argparse import Namespace
import logging as log
import time


PUT_TIMEOUT = 30
GET_TIMEOUT = 150
QUERY_THREAD_TIMEOUT = 10
QUERY_THREAD_IDLE_TIMEOUT = 60
JOIN_DYNAMIC_SLEEP_TIME_INCREMENT = 10

def _parse_date_andco(item):
    """A custom filter to transform JSON date (i.e. date in ISO formated text) 
    in datetime object. JSON object can be list of dicts or a simple dict or
    a simple object directly in some case. Only variable with name ending with 
    _date are converted (thus only within a dictionary, standalone object are
    not converted)
    Perform also misc transformation like batch default"""
    def _sub_parse(subitem):
        for key, value in subitem.items():
            if '_date' in key and value is not None:
                subitem[key] = datetime.fromisoformat(value)
            elif key=='batch' and value is None:
                subitem[key] = 'Default'
        return subitem
    if type(item)==list:
        return list([_sub_parse(subitem) for subitem in item])
    elif type(item)==dict:
        return _sub_parse(item)
    else:
        return item

def _clean(d):
    """filter out null value from a dict as well as key in remove list (default
    to ['id']) plus a ugly hack for default batch"""
    return dict([('batch',None) if (k=='batch' and v=='Default') else (k,v)  
        for k,v in d.items() if v is not None])

def _to_obj(d):
    """A simple recursive wrapper to convert JSON like objects in python objects"""
    if type(d)==list:
        return map(_to_obj, d)
    if type(d)==dict:
        return Namespace(**d)
    return d

class HTTPException(Exception):
    def __init__(self, text, message, status_code):
        self.status_code = status_code
        self.message = message
        super().__init__(text)


def _filter_non200(r):
    if r.status_code//100 != 2:
        try:
            message = r.json()['message']
        except:
            message = 'Unknown server error'
        raise HTTPException(f'Error {r.status_code}: {message}', 
                message=message, 
                status_code=r.status_code)
    return r.json()
class LazyObject:
    """A lazy object or dict initialized with a queue.Queue 
    perform q.get as soon as you try to use it..."""
    def __init__(self, q):
        """Initialize object with queue.Queue q"""
        self._q = q
        self._o = None
    def __getattr__(self, __name):
        """Retrieve real attribute (object style) in real object once it is there"""
        if self._o is None:
            self._o = self._q.get()
        return getattr(self._o, __name)
    def __getitem__(self, __key):
        """Retrieve real key in real object (dict style) once it is there"""
        if self._o is None:
            self._o = self._q.get()
        return self._o[__key]
    def __str__(self):
        """Wait for real object and use real object representation"""
        if self._o is None:
            self._o = self._q.get()
        return str(self._o)


def query_thread(send_queue, put_timeout):
        """A Query-thread which treats all the queries with a while loop until the object is there, then it put the result in the returning queue
        """
        time_point = None
        while True :
            try:
                self,type,query,return_queue = send_queue.get(timeout=QUERY_THREAD_TIMEOUT)
                time_point = None
            except queue.Empty:
                if time_point is None:
                    time_point=time.time()
                elif time.time() - time_point > QUERY_THREAD_IDLE_TIMEOUT:
                    return None
                sleep(put_timeout)
                continue
            url,data=query
            
            task_done=False
            while not task_done:
                try:
                    if type=='put':
                        result = self._wrap(requests.put(
                                self.url+url, json=data, timeout=put_timeout))
                    elif type=='post':
                        result = self._wrap(requests.post(
                                self.url+url, json=data, timeout=put_timeout))
                    elif type=='delete':
                        result = self._wrap(requests.delete(
                                self.url+url, timeout=put_timeout))
                    task_done = True
                except (ConnectionError,Timeout) as e:
                    log.exception(f'Exception when trying to {type}: {e}')
            return_queue.put(result)

class RestartingThread:
    """This Thread class is a thin wrapper above a threading.Thread that can be 
    restarted, it implements only start() and is_alive() methods
    """
    def __init__(self, target, args):
        self.target = target
        self.args = args
        self.__thread__=None
    
    def is_alive(self):
        return False

    def start(self):
        if self.is_alive():
            print('Thread already alive')
            return
        self.__thread__=threading.Thread(target=self.target, args=self.args)
        self.__thread__.start()
        self.is_alive = self.__thread__.is_alive
        


class Server:
    """A thin wrapper above requests with URLs registered
    and several extrathings:
    - output is JSON de-marshalled,
    - date are converted,
    - if server is not accessible, the lib will retry or do the job later,
    - expected arguments are explicit which makes life easier.
    
    Arguments:
    - ip: the name or IP address of server,
    - style: if 'dict' (default) all the objects are rendered as dict, if 
        'object', dict are passed to Namespace to produce real Python objects.
    """

    def __init__(self, ip, style='dict', asynchronous=True, 
            put_timeout=PUT_TIMEOUT, get_timeout=GET_TIMEOUT):
        """initialise the object with IP or name of the server
        
        Arguments:
        - ip: the name or IP address of server,
        - style: if 'dict' (default) all the objects are rendered as dict, if 
            'object', dict are passed to Namespace to produce real Python objects.
        - asynchronous: if True (default) then put and post operations are 
            asynchronous, if False then all operations will wait (forever) for 
            the server to come back. get operations are always synchronous"""
        self.ip=ip
        self.url=f'http://{ip}:5000'
        if style=='dict':
            self._wrap = lambda x: _parse_date_andco(_filter_non200(x))
        elif style=='object':
            self._wrap = lambda x: _to_obj(_parse_date_andco(_filter_non200(x)))
        else:
            raise Exception(f'style can only be dict or object not f{style}')
        self.asynchronous = asynchronous
        if self.asynchronous:
            self.send_queue= queue.Queue()
            self.query_thread=RestartingThread(target=query_thread,
                args=(self.send_queue,put_timeout))
        self.get_timeout = get_timeout
        self.put_timeout = put_timeout

    def queue_size(self):
        """A method to estimate send queue size"""
        return self.send_queue.qsize()        

    def get(self, url, **args):
        """A wrapper used for all get operations.
        - url: extra string to add after base server URL
        
        return the objects according to Server style (see style in class doc)"""
        try:
            return self._wrap(requests.get(
                self.url+url, timeout=self.get_timeout, json=args
            ))
        except (ConnectionError,Timeout) as e:
            log.warning(f'Exception when trying to get: {e}')
            sleep(self.get_timeout)
            return self.get(url, **args)

    def put(self,url, data, asynchronous=None, timeout=None):
        """A wrapper used for all put operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of put operation) represented as dict

        return the objects according to Server style (see style in class doc)
        (in asynchronous mode, return a lazy object)"""
        asynchronous = self.asynchronous if asynchronous is None else asynchronous
        if timeout is None:
            timeout = self.put_timeout if asynchronous else self.get_timeout
        try:
            return self._wrap(requests.put(
                self.url+url, json=data, timeout=timeout
            ))
        except (ConnectionError,Timeout) as e:
            if asynchronous:
                if not self.query_thread.is_alive():
                    self.query_thread.start()
                log.warning(f'Exception when trying to put: {e}')
                return_queue=queue.Queue()
                self.send_queue.put((self,'put',(url, data),return_queue))
                return LazyObject(return_queue)
            else:
                log.warning(f'Exception when trying to put: {e}')
                sleep(timeout)
                return self.put(url, data, asynchronous, timeout)
        

    def post(self, url, data, asynchronous=None, timeout=None):
        """A wrapper used for all post operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of post operation) represented as dict

        return the objects according to Server style (see style in class doc)"""
        asynchronous = self.asynchronous if asynchronous is None else asynchronous
        if timeout is None:
            timeout = self.put_timeout if asynchronous else self.get_timeout
        try:
            return self._wrap(requests.post(
                url=self.url+url, json=data, timeout=self.put_timeout
            ))
        except (ConnectionError,Timeout) as e:
            if asynchronous:
                if not self.query_thread.is_alive():
                    self.query_thread.start()
                log.warning(f'Exception when trying to post: {e}')
                return_queue=queue.Queue()
                self.send_queue.put((self,'post',(url, data),return_queue))
                return LazyObject(return_queue)
            else:
                log.warning(f'Exception when trying to post: {e}')
                sleep(timeout)
                return self.post(url, data, asynchronous, timeout)

    def delete(self,url, asynchronous=None, timeout=None):
        """A wrapper used for all put operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of put operation) represented as dict

        return the objects according to Server style (see style in class doc)
        (in asynchronous mode, return a lazy object)"""
        asynchronous = self.asynchronous if asynchronous is None else asynchronous
        if timeout is None:
            timeout = self.put_timeout if asynchronous else self.get_timeout
        try:
            return self._wrap(requests.delete(
                self.url+url, timeout=timeout
            ))
        except (ConnectionError,Timeout) as e:
            if asynchronous:
                if not self.query_thread.is_alive():
                    self.query_thread.start()
                log.warning(f'Exception when trying to delete: {e}')
                return_queue=queue.Queue()
                self.send_queue.put((self,'delete',(url, None),return_queue))
                return LazyObject(return_queue)
            else:
                log.warning(f'Exception when trying to delete: {e}')
                sleep(timeout)
                return self.delete(url, asynchronous, timeout)


    def workers(self):
        """Get a list of all workers"""
        return self.get('/workers/')

    def worker_update(self, id, name=None, hostname=None, concurrency=None, 
                    status=None,batch=None, idle_callback=None, prefetch=None,
                    asynchronous=True):
        """Update a specific worker with worker_id equal to id
        return the updated worker (or None if the server timeout)"""
        return self.put(f'/workers/{id}', data=_clean(
            {'name':name, 'hostname':hostname, 'concurrency':concurrency, 
            'status':status, 'batch':batch, 'idle_callback':idle_callback,
            'prefetch':prefetch}
        ), asynchronous=asynchronous)

    def worker_create(self, name, concurrency, hostname=None, status='paused',
                batch=None, idle_callback=None, prefetch=0, asynchronous=True):
        """Create a new worker
        return the new worker (or None if the server timeout)"""
        return self.post('/workers/', data=_clean(
            {'name':name, 'hostname':hostname, 'concurrency':concurrency, 
            'status':status, 'batch':batch, 'idle_callback':idle_callback,
            'prefetch':prefetch}
        ), asynchronous=asynchronous)

    def worker_get(self, id):
        """get a specific worker with worker_id equal to id
        return the worker"""
        return self.get(f'/workers/{id}')

    def worker_ping(self, id, load, memory, stats, asynchronous=False):
        """Update a specific worker ping time (heartbit)
        return a updated worker object with attribute or key (depending on style)"""
        return self.put(f'/workers/{id}/ping', data={'load':load,'memory':memory,
            'stats':stats}, asynchronous=asynchronous)

    def worker_callback(self, id, message, asynchronous=False):
        """Send a callback message (mainly idle) to trigger action on worker from the server
        return a object with result attribute, equal to ok"""
        return self.put(f'/workers/{id}/callback', data={'message':message},
            asynchronous=asynchronous)
    
    def worker_executions(self, id, status=None):
        """Get a list of a worker current assigned executions. 
        If status is used, only executions with this status are shown
        """
        if status is None:
            return self.get(f'/workers/{id}/executions')
        else:
            return self.get(f'/workers/{id}/executions/{status}')

    def worker_delete(self, id, asynchronous=True):
        """Delete a worker"""
        return self.delete(f'/workers/{id}', asynchronous=asynchronous)

    def workers_tasks(self):
        """Return a list of all the different tasks for a certain worker"""
        return self.get(f'/workers/tasks')
    
    def executions(self, **args):
        """Get a list of all executions. 
        Optional args are optional attributes to filter list with:
        - task_id: int
        - status: str
        - latest: boolean
        """
        return self.get(f'/executions/', **args)
    
    def execution_create(self, worker_id, task_id, status='pending', command=None,
                         asynchronous=True):
        """Create a new execution, return the newly created execution
        """
        return self.post('/executions/', data=_clean(
            {
            'worker_id':worker_id, 'task_id':task_id, 'command':command,
            'status':status
            }
            ), asynchronous=asynchronous)

    def execution_update(self, id, status=None, pid=None, return_code=None, 
                        output=None, output_files=None, command=None,
                        asynchronous=True):
        """Update a specific execution, return the updated execution
        """
        return self.put(f'/executions/{id}', data=_clean(
            {'status':status, 'pid':pid, 'return_code':return_code, 
                'output':output, 'output_files':output_files, 'command':command}
        ), asynchronous=asynchronous)

    def execution_output_write(self, id, output, asynchronous=True):
        """Add some output to a specific execution, return the updated execution
        """
        if output is None:
            return None
        output = output.replace('\00', '')
        return self.put(f'/executions/{id}/output', data=_clean(
            {'text':output}
        ), asynchronous=asynchronous)

    def execution_error_write(self, id, error, asynchronous=True):
        """Add some error (stderr) to a specific execution, return the updated execution
        """
        if error is None:
            return None
        error = error.replace('\00', '')
        return self.put(f'/executions/{id}/error', data=_clean(
            {'text':error}
        ), asynchronous=asynchronous)

    def execution_get(self, id):
        """get a specific execution with execution_id equal to id
        return the execution"""
        return self.get(f'/executions/{id}')

    def task_create(self, command, name=None, status='pending',batch=None, 
            input=None, output=None, container=None, container_options='',
            resource=None, asynchronous=True):
        """Create a new task, return the newly created task
        """
        return self.post('/tasks/', data=_clean({
            'command':command, 'name':name, 'status':status, 'batch':batch,
            'input':input, 'output':output, 'container':container, 
            'container_options':container_options, 'resource':resource
        }), asynchronous=asynchronous)

    def task_update(self, id, command=None, name=None, status=None, batch=None, 
            input=None, output=None, container=None, container_options=None,
            resource=None, asynchronous=True):
        """Update a specific execution, return the updated execution
        """
        return self.put(f'/tasks/{id}', data=_clean({
            'command':command, 'name':name, 'status':status, 'batch':batch,
            'input':input, 'output':output, 'container':container, 
            'container_options':container_options, 'resource':resource
        }), asynchronous=asynchronous)

    def task_get(self, id):
        """get a specific task with task_id equal to id
        return the task"""
        return self.get(f'/tasks/{id}')

    def tasks(self, **args):
        """Get a list of all tasks
        Optional args are optional attributes to filter list with:
        - task_id: list(int)
        - status: str
        - batch: str
        - name: str
        """
        return self.get(f'/tasks/', **args)

    def task_delete(self,id, asynchronous=True):
        """delete a specific task"""
        return self.delete(f'/tasks/{id}', asynchronous=asynchronous)

    def worker_signals(self, id):
        """Get a list of signals for this worker"""
        return self.get(f'/workers/{id}/signals')

    def worker_create_signal(self, id, execution_id, signal, asynchronous=True):
        """Create a signal for this worker"""
        return self.post(f'/workers/{id}/signals', data={
            'execution_id': execution_id,
            'signal': signal
        }, asynchronous=asynchronous)
    
    def worker_deploy(self, number, batch, region, flavor, concurrency, provider,
                      prefetch=0,
             asynchronous=True):
        """Deploy (and create) some workers"""
        return self.put('/workers/deploy', data={
            'number':int(number),
            'region':region,
            'provider':provider,
            'flavor':flavor,
            'batch':batch,
            'concurrency':concurrency,
            'prefetch': prefetch
        }, asynchronous=asynchronous)
    
    def batch_stop(self,batch,signal=0, asynchronous=True):
        """Pause all workers for this batch"""
        if signal==0:
            return self.put(f'/batch/{batch}/stop',data={})
        return self.put(f'/batch/{batch}/stop',data={'signal':signal}, 
            asynchronous=asynchronous)
    
    def batch_go(self,batch,signal=0, asynchronous=True):
        """(re)set all workers affected to this batch to running"""
        return self.put(f'/batch/{batch}/go',data={'signal':signal},
            asynchronous=asynchronous)

    def batches(self):
        """List all batches, their tasks and workers"""
        return self.get(f'/batch/')

    def batch_delete(self,batch, asynchronous=True):
        """List all batches, their tasks and workers"""
        return self.delete(f'/batch/{batch}', asynchronous=asynchronous)

    def join(self, task_list, retry=1, check=False):
        """Wait for a certain list of tasks to succeed. In case of failure, relaunch tasks
        a limited number of time (retry). If check is True, then join will fail if
        one of the task fails (after all retries).
        Return a dictionary of the different status of tasks when tasks were all done
        """
        if not task_list:
            raise RuntimeError('task_list should not be empty')
        task_ids = [task['task_id'] if type(task)==dict else task.task_id for task in task_list]
        task_retries = dict([(task_id,0) for task_id in task_ids])

        all_task_done = False
        tasks = None
        dynamic_sleep_time = True
        sleep_time = JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
        failed_tasks = []
        while not all_task_done:
            all_task_done = True
            old_tasks = tasks
            tasks = {'assigned':0,'failed':0, 'running':0, 'accepted':0, 'pending':0,'succeeded':0, 'paused':0}
            for task in self.tasks(task_id=task_ids):
                if type(task)==dict:
                    task = Namespace(**task)
                if task.status=='failed':
                    if task_retries[task.task_id]<retry:
                        print(f'Retrying task {task.name or task.task_id} [{task_retries[task.task_id]+1}/{retry}]...')
                        self.task_update(task.task_id, status='pending')
                        task_retries[task.task_id]+=1
                        tasks['pending']+=1
                        all_task_done = False
                    else:
                        tasks['failed']+=1
                        if task.task_id not in failed_tasks:
                            print(f'Task {task.name or task.task_id} failed too many times giving up')
                            failed_tasks.append(task.task_id)
                elif task.status in ['running','accepted','pending','assigned']:
                    all_task_done = False
                    tasks[task.status]+=1
                else:
                    tasks[task.status]+=1
            print(f"Remaining tasks pending : {tasks['pending']}, assigned: {tasks['assigned']}, accepted: {tasks['accepted']}, running: {tasks['running']}, failed: {tasks['failed']}, succeeded: {tasks['succeeded']}")
            
            # sleeping if needed
            if not all_task_done:
                sleep(sleep_time)

            # adjust sleeping time
            if old_tasks is not None and dynamic_sleep_time:
                if tasks['succeeded']==old_tasks['succeeded'] and tasks['failed']==old_tasks['failed']:
                    # no task has been completed since last iteration so it is likely that the time loop is two short
                    sleep_time += JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
                else:
                    # ok so now some tasks have been completed so it is likely the time loop is correct, no need to move it further
                    dynamic_sleep_time = False
            elif not dynamic_sleep_time and sleep_time>JOIN_DYNAMIC_SLEEP_TIME_INCREMENT:
                if tasks['succeeded']!=old_tasks['succeeded'] or tasks['failed']!=old_tasks['failed']:
                    # after all it seems it is moving quicker than expected, maybe time to shorten the time loop
                    dynamic_sleep_time = True
                    sleep_time -= JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
        if check and tasks['failed']>0:
            raise RuntimeError('Could not complete all the tasks...')
        print('All tasks done!')
        return tasks


if __name__=="__main__":
    s=Server('127.0.0.1',style='object')



import botocore.exceptions
from ftplib import FTP
import re
import logging as log
from functools import wraps, cached_property
from time import sleep
import shutil
import os
import requests
import glob
import hashlib
import subprocess
from .util import PropagatingThread, xboto3, if_is_not_None
import concurrent.futures
import argparse
import datetime
import pytz
import io
from azure.storage.blob import BlobServiceClient, BlobBlock
import uuid
import azure.core.exceptions
from .constants import DEFAULT_WORKER_CONF
import dotenv
from tabulate import tabulate 
from fnmatch import fnmatch

# how many time do we retry
RETRY_TIME = 3
RETRY_SLEEP_TIME = 10
MAX_SLEEP_TIME = 900
PUBLIC_RETRY_TIME = 20
MAX_PARALLEL_S3 = 5

# position in FTP dir command
# typical output is: '-rw-rw-r--    1 ftp      ftp       1088322 Jul 15  2019 MGYG-HGUT-00001.faa' 
FTP_DIR_SIZE_POSITION = 4
FTP_DIR_MONTH_POSITION = 5
FTP_DIR_DAY_POSITION = 6
FTP_DIR_YEAR_POSITION = 7
FTP_DIR_NAME_POSITION = 8

# name of Azure variables
AZURE_ACCOUNT='SCITQ_AZURE_ACCOUNT'
AZURE_KEY='SCITQ_AZURE_KEY'
MAX_PARALLEL_AZURE = 5
AZURE_CHUNK_SIZE = 50*1024**2

MAX_PARALLEL_SYNC = 10

class FetchError(Exception):
    pass

def pathjoin(*items):
    """Like os.path.join but always with '/' even with Windows and no double '/'
        in the middle of the path 
        (also no funny things if one of the item start with slash)
        (multiple slashes will be kept if at the begining or the end)"""
    last = len(items) - 1 
    joined_path = '/'.join([ item.rstrip('/') if rank==0 else 
                      item.lstrip('/') if rank==last else 
                      item.strip('/') 
                      for rank,item in enumerate(items)])
    # s3 tend to do funny things with that pattern, let us remove that buggy path
    joined_path = joined_path.replace('/./','/')
    return joined_path

# general
def retry_if_it_fails(n):
    """A decorator to retry n time some action"""
    def decorator(function):
        @wraps(function)
        def wrapper(*args, __retry_number__=None, **kwargs):
            iteration=0
            if __retry_number__ is None:
                __retry_number__ = n
            sleep_time=RETRY_SLEEP_TIME
            while iteration<__retry_number__:
                try:
                    retval = function(*args, **kwargs)
                    break
                except Exception:
                    log.exception('Something went bad')
                    iteration += 1
                    if iteration<__retry_number__:
                        log.warning(f'Waiting some time ({sleep_time}s)...')
                        sleep(sleep_time)
                        if sleep_time < MAX_SLEEP_TIME/2:
                            sleep_time *= 2
                        else:
                            sleep_time = MAX_SLEEP_TIME
                        log.warning('Retrying...')
                    else:
                        log.error('Too many failures, giving up')
                        raise
            return retval
        return wrapper
    return decorator    



def complete_if_ends_with_slash(source, destination):
    """This function checks if destination ends with slash in which case it completes
    with the source last item
    """
    if destination.endswith('/'):
        destination += source.split('/')[-1]
    return destination

# actions 

def gunzip(filepath):
    """Stupid gunzipper with gzip"""
    subprocess.run(['pigz','-d',filepath], check=True)
    
def untar(filepath):
    """Untar the tar archive locally where it is and delete the archive.
    (so it behaves like gunzip, and not like tar usually)"""
    path, basename = os.path.split(filepath)
    if basename.endswith('gz'):
        subprocess.run(f'pigz -dc "{basename}"|tar x', cwd=path, shell=True, check=True)
    else:
        subprocess.run(['tar','xf',basename], cwd=path, check=True)
    os.remove(filepath)

# AWS S3 

S3_REGEXP=re.compile(r'^s3://(?P<bucket>[^/]*)/(?P<path>.*)$')

def get_s3():
    """Replace boto3.resource('s3') using hack suggested in https://github.com/aws/aws-cli/issues/1270"""
    return xboto3().resource('s3')

def _bucket_get(bucket, key, destination):
    """A small helper just to ease boto3 client in thread/process context to download a file"""
    return get_s3().Bucket(bucket).download_file(key, destination)

@retry_if_it_fails(RETRY_TIME)
def s3_get(source, destination):
    """S3 downloader: download source expressed as s3://bucket/path_to_file 
    to destination - a local file path"""
    log.info(f'S3 downloading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = S3_REGEXP.match(source).groupdict()
    if uri_match['path'].endswith('/'):
        retry = RETRY_TIME
        try:
            bucket=get_s3().Bucket(uri_match['bucket'])
            objects = list(bucket.objects.filter(Prefix=uri_match['path']))
            while retry>0:
                failed_objects = []
                failed = False
                try:
                    jobs = {}
                    with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_PARALLEL_S3) as executor:
                        for obj in objects:
                            destination_name = os.path.relpath(obj.key, uri_match['path'])
                            destination_name = os.path.join(destination, destination_name)
                            destination_path,_ = os.path.split(destination_name)
                            log.info(f'S3 downloading {obj.key} to {destination_name}')
                            if not os.path.exists(destination_path):
                                os.makedirs(destination_path, exist_ok=True)
                            if not os.path.exists(destination_name):
                                jobs[executor.submit(_bucket_get, uri_match['bucket'], obj.key, destination_name)]=obj
                        for job in  concurrent.futures.as_completed(jobs):
                            obj = jobs[job]
                            if job.exception() is None:
                                log.info(f'Done for {obj.key}: {job.result()}')
                            else:
                                log.error(f'Could not download {obj.key}')
                                log.exception(job.exception())
                                failed = True
                                failed_objects.append(obj)
                    if failed:
                        objects = failed_objects
                        retry -= 1
                        continue
                    else:
                        break
                except:
                    pass
        except botocore.exceptions.ClientError as error:
            if 'Not Found' in error.response.get('Error',{}).get('Message',None):
                raise FetchError(f'{source} was not found') from error
            else:
                raise
        
        if failed:
            raise FetchError(f"These objects could not be downloaded: {','.join([obj.key for obj in failed_objects])}")
    else:
        try:
            xboto3().client('s3').download_file(uri_match['bucket'],
                uri_match['path'],destination)
        except botocore.exceptions.ClientError as error:
            if 'Not Found' in error.response.get('Error',{}).get('Message',None):
                raise FetchError(f'{source} was not found') from error
            else:
                raise


@retry_if_it_fails(RETRY_TIME)
def s3_put(source, destination):
    """S3 uploader: download a local file path in source to a destination
    expressed as a s3 URI s3://bucket/path_to_file"""
    log.info(f'S3 uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = S3_REGEXP.match(destination).groupdict()
    xboto3().client('s3').upload_file(source,uri_match['bucket'],
        uri_match['path'])

def s3_info(uri):
    """S3 info fetcher: get some info on a blob specified with s3://bucket/path_to_file"""
    log.info(f'S3 getting info for {uri}')
    uri_match = S3_REGEXP.match(uri).groupdict()
    try:
        bucket=get_s3().Bucket(uri_match['bucket'])
        object = next(iter(bucket.objects.filter(Prefix=uri_match['path'])))
        return argparse.Namespace(size=object.size, 
                                  creation_date=object.last_modified,
                                  modification_date=object.last_modified)
    except StopIteration:
        raise FetchError(f'{uri} was not found')

def s3_list(uri, no_rec = False):
    """List content of an s3 folder in the form s3://bucket/path
    if no_rec is True, restrict listing to path (and not recursive) - which may list "folders" with None creation_date"""
    log.info(f'S3 getting listing for {uri}')
    uri_match = S3_REGEXP.match(uri).groupdict()
    if no_rec:
        s3_client=xboto3().client('s3')
        bucket_name=uri_match['bucket']
        query = s3_client.list_objects_v2(
                Bucket=bucket_name,
                Delimiter='/',
                Prefix=uri_match['path'])
        return [argparse.Namespace(name=f's3://{bucket_name}/{prefix["Prefix"]}',
                                   rel_name=os.path.relpath(prefix['Prefix'], uri_match['path'])+'/',
                                   size=0,
                                   creation_date=None,
                                   modification_date=None) for prefix in query['CommonPrefixes']
                ] if 'CommonPrefixes' in query else [] + [
                argparse.Namespace(name=f's3://{bucket_name}/{object["Key"]}',
                            rel_name=os.path.relpath(object['Key'], uri_match['path']),
                            size=object['Size'], 
                            creation_date=object['LastModified'],
                            modification_date=object['LastModified'])
                for object in query['Contents']] if 'Contents' in query else []
    else:
        bucket=get_s3().Bucket(uri_match['bucket'])
        return [argparse.Namespace(name=f's3://{bucket.name}/{object.key}',
                                rel_name=os.path.relpath(object.key, uri_match['path']),
                                size=object.size, 
                                creation_date=object.last_modified,
                                modification_date=object.last_modified)
                for object in bucket.objects.filter(Prefix=uri_match['path'])]

def s3_delete(uri):
    """Delete an s3 blob"""
    log.info(f'S3 deleting {uri}')
    uri_match = S3_REGEXP.match(uri).groupdict()
    xboto3().client('s3').delete_object(Bucket=uri_match['bucket'], Key=uri_match['path'])


# Azure

AZURE_REGEXP=re.compile(r'^azure://(?P<container>[^/]*)/(?P<path>.*)$')


def _container_get(container, blob_name, file_name):
    """A small wrapper to ease azure client in thread/process context to download a file
    container can be either a string or an Azure container client."""
    if type(container)==str:
        container = AzureClient().client.get_container_client(container)
    with open(file=file_name, mode="wb") as download_file:
        download_file.write(container.download_blob(blob_name).readall())

class AzureClient:
    """A small wrapper above Azure blob client to integrate with scitq"""
    def __init__(self):
        """This deals with initialization"""
        account_name=os.environ.get(AZURE_ACCOUNT)
        account_key=os.environ.get(AZURE_KEY)
        if account_name is None:
            if os.path.isfile(DEFAULT_WORKER_CONF):
                dotenv.load_dotenv(DEFAULT_WORKER_CONF)
                account_name=os.environ.get(AZURE_ACCOUNT)
                account_key=os.environ.get(AZURE_KEY)
        if account_name is None:
            raise FetchError(f'Azure account is not properly configured: either set {AZURE_ACCOUNT} and {AZURE_KEY} environment variables or adjust {DEFAULT_WORKER_CONF}.')
        connection_string = f"DefaultEndpointsProtocol=https;AccountName={account_name};AccountKey={account_key};EndpointSuffix=core.windows.net"
        self.client = BlobServiceClient.from_connection_string(connection_string)

    @retry_if_it_fails(RETRY_TIME)
    def get(self, source, destination):
        log.info(f'Azure downloading {source} to {destination}')
        destination=complete_if_ends_with_slash(source, destination)
        uri_match = AZURE_REGEXP.match(source).groupdict()
        if uri_match['path'].endswith('/'):
            retry = RETRY_TIME
            try:
                container=self.client.get_container_client(uri_match['container'])
                objects = list(container.list_blobs(name_starts_with=uri_match['path']))
                while retry>0:
                    failed_objects = []
                    failed = False
                    try:
                        jobs = {}
                        with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_PARALLEL_AZURE) as executor:
                            for obj in objects:
                                destination_name = os.path.relpath(obj.name, uri_match['path'])
                                destination_name = os.path.join(destination, destination_name)
                                destination_path,_ = os.path.split(destination_name)
                                log.info(f'Azure downloading {obj.name} to {destination_name}')
                                if not os.path.exists(destination_path):
                                    os.makedirs(destination_path, exist_ok=True)
                                if not os.path.exists(destination_name):
                                    jobs[executor.submit(_container_get, 
                                            container.container_name, 
                                            obj.name, 
                                            destination_name)]=obj
                            for job in  concurrent.futures.as_completed(jobs):
                                obj = jobs[job]
                                if job.exception() is None:
                                    log.info(f'Done for {obj.key}: {job.result()}')
                                else:
                                    log.error(f'Could not download {obj.key}')
                                    log.exception(job.exception())
                                    failed = True
                                    failed_objects.append(obj)
                        if failed:
                            objects = failed_objects
                            retry -= 1
                            continue
                        else:
                            break
                    except:
                        pass
            except azure.core.exceptions.ResourceNotFoundError as error:
                raise FetchError(f'{source} was not found') from error
            
            if failed:
                raise FetchError(f"These objects could not be downloaded: {','.join([obj.key for obj in failed_objects])}")
        else:
            try:
                container=self.client.get_container_client(uri_match['container'])
                _container_get(container, uri_match['path'],destination)
            except azure.core.exceptions.ResourceNotFoundError as error:
                raise FetchError(f'{source} was not found') from error

    @retry_if_it_fails(RETRY_TIME)
    def put(self, source, destination):
        """Azure uploader: download a local file path in source to a destination
        expressed as an Azure URI azure://container/path_to_file"""
        log.info(f'Azure uploading {source} to {destination}')
        destination=complete_if_ends_with_slash(source, destination)
        uri_match = AZURE_REGEXP.match(destination).groupdict()
        blob_client = self.client.get_blob_client(container=uri_match['container'],
                                                blob=uri_match['path'])
        with open(file=source, mode="rb") as data:
            if os.path.getsize(source)>=AZURE_CHUNK_SIZE:
                if blob_client.exists():
                    blob_client.delete_blob()
                block_list=[]
                while True:
                    read_data = data.read(AZURE_CHUNK_SIZE)
                    if not read_data:
                        break # done
                    blk_id = str(uuid.uuid4())
                    blob_client.stage_block(block_id=blk_id,data=read_data) 
                    block_list.append(BlobBlock(block_id=blk_id))
                blob_client.commit_block_list(block_list)
            else:
                blob_client.upload_blob(data, overwrite=True)

    def info(self,uri):
        """Azure info fetcher: get some info on a blob specified with azure://container/path_to_file"""
        log.info(f'Azure getting info for {uri}')
        uri_match = AZURE_REGEXP.match(uri).groupdict()
        try:
            container=self.client.get_container_client(uri_match['container'])
            object = next(iter(container.list_blobs(name_starts_with=uri_match['path'])))
            return argparse.Namespace(size=object.size, 
                                    creation_date=object.creation_time,
                                    modification_date=object.last_modified)
        except StopIteration:
            raise FetchError(f'{uri} was not found')

    def list(self,uri, no_rec=False):
        """List the content of an azure storage folder  azure://container/path
        if no_rec is True, restrict listing to path (and not recursive) - which may list "folders" with None creation_date"""
        log.info(f'Azure getting listing for {uri}')
        uri_match = AZURE_REGEXP.match(uri).groupdict()
        container=self.client.get_container_client(uri_match['container'])
        if no_rec:
            return [argparse.Namespace(name=f"azure://{container.container_name}/{object.name}",
                                    rel_name=os.path.relpath(object.name, uri_match['path'])+('/' if object.name.endswith('/') else ''),
                                    size=object.get('size',0), 
                                    creation_date=object.get('creation_time',None),
                                    modification_date=object.get('last_modified',None))
                    for object in container.walk_blobs(name_starts_with=uri_match['path'],delimiter='/')]
        else:
            return [argparse.Namespace(name=f"azure://{container.container_name}/{object.name}",
                                    rel_name=os.path.relpath(object.name, uri_match['path']),
                                    size=object.size, 
                                    creation_date=object.creation_time,
                                    modification_date=object.last_modified)
                    for object in container.list_blobs(name_starts_with=uri_match['path'])]

    def delete(self, uri):
        """Delete an azure blob"""
        log.info(f'Azure deleting {uri}')
        uri_match = AZURE_REGEXP.match(uri).groupdict()
        container=self.client.get_container_client(uri_match['container'])
        container.delete_blob(uri_match['path'])


# FTP

FTP_REGEXP=re.compile(r'^ftp://(?P<host>[^/]*)/(?P<path>.*)$')

@retry_if_it_fails(RETRY_TIME)
def ftp_get(source, destination):
    """FTP downloader: download source expressed as ftp://host/path_to_file 
    to destination - a local file path"""
    log.info(f'FTP downloading {source} to {destination}')
    if source.endswith('/'):
        raise FetchError('Directory fetching is not yet supported for ftp:// URI')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FTP_REGEXP.match(source)
    if not uri_match:
        message=f'Source does not seem a proper FTP URL: {source}'
        log.error(message)
        raise FetchError(message)
    uri_match = uri_match.groupdict()
    with open(destination, 'wb') as local_file:
        with FTP(uri_match['host']) as ftp:
            ftp.login()
            ftp.retrbinary(f"RETR {uri_match['path']}", local_file.write)

@retry_if_it_fails(RETRY_TIME)
def ftp_put(source, destination):
    """FTP uploader: upload source expressed as a local file path 
    to destination expressed as a FTP URL: ftp://host/path_to_file """
    log.info(f'FTP uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FTP_REGEXP.match(destination).groupdict()
    with open(destination, 'rb') as local_file:
        with FTP(uri_match['host']) as ftp:
            ftp.login()
            ftp.storbinary(f"STOR {uri_match['path']}", local_file.read)

def ftp_info(uri):
    """FTP info: get some information on a FTP object"""
    log.info(f'FTP get info for {uri}')
    uri_match = FTP_REGEXP.match(uri).groupdict()
    with FTP(uri_match['host']) as ftp:
        ftp.login()

        with io.StringIO() as output:
            ftp.dir(uri_match['path'],output.write)
            output.seek(0)
            answer = output.read()
        
        if answer:
            obj = answer.split()
            obj_date = datetime.datetime.strptime(
                ' '.join( (obj[FTP_DIR_MONTH_POSITION],
                           obj[FTP_DIR_DAY_POSITION],
                           obj[FTP_DIR_YEAR_POSITION])
                    ),
                '%b %d %Y').replace(tzinfo=pytz.utc)
            return argparse.Namespace(size=obj[FTP_DIR_SIZE_POSITION],
                                    creation_date=obj_date, 
                                    modification_date=obj_date)
        else:
            raise FetchError(f'{uri} was not found')

def ftp_list(uri, no_rec=False):
    """list recursively the content of a FTP folder
    not recursively if no_rec is True"""
    log.info(f'FTP get info for {uri}')
    uri_match = FTP_REGEXP.match(uri).groupdict()
    answer = []
    with FTP(uri_match['host']) as ftp:
        ftp.login()

        with io.StringIO() as output:
            listing = []
            ftp.dir(uri_match['path'],listing.append)
        
        for item in listing:
            obj = item.split()
            is_dir=item[0]=='d'
            if is_dir:
                if no_rec:
                    try:
                        obj_date = datetime.datetime.strptime(
                            ' '.join( (obj[FTP_DIR_MONTH_POSITION],
                                    obj[FTP_DIR_DAY_POSITION],
                                    obj[FTP_DIR_YEAR_POSITION])
                                ),
                            '%b %d %Y').replace(tzinfo=pytz.utc)
                    except ValueError:
                        try:
                            obj_date = datetime.datetime.strptime(
                                ' '.join( (obj[FTP_DIR_MONTH_POSITION],
                                        obj[FTP_DIR_DAY_POSITION],
                                        str(datetime.datetime.now().year))
                                    ),
                                '%b %d %Y').replace(tzinfo=pytz.utc)
                        except ValueError:
                            obj_date = None
                    answer.append(argparse.Namespace(
                        name=f"ftp://{uri_match['host']}/{os.path.join(uri_match['path'],obj[FTP_DIR_NAME_POSITION]+'/')}",
                        rel_name=obj[FTP_DIR_NAME_POSITION]+'/',
                        size=0,
                        creation_date = obj_date,
                        modification_date=obj_date
                    ))
                else:
                    dir_name = obj[FTP_DIR_NAME_POSITION]
                    for item in ftp_list(os.path.join(uri, dir_name)):
                        item.rel_name = f"{dir_name}/{item.rel_name}"
                        answer.append(item)
            else:
                obj_date = datetime.datetime.strptime(
                    ' '.join( (obj[FTP_DIR_MONTH_POSITION],
                            obj[FTP_DIR_DAY_POSITION],
                            obj[FTP_DIR_YEAR_POSITION])
                        ),
                    '%b %d %Y').replace(tzinfo=pytz.utc)
                answer.append(argparse.Namespace(
                    name=f"ftp://{uri_match['host']}/{os.path.join(uri_match['path'],obj[FTP_DIR_NAME_POSITION])}",
                    rel_name=obj[FTP_DIR_NAME_POSITION],
                    size=obj[FTP_DIR_SIZE_POSITION],
                    creation_date=obj_date, 
                    modification_date=obj_date))

    return answer

@cached_property
def docker_available():
    """A property to see if docker is there"""
    return subprocess.run(['docker', '-v'],check=False).returncode==0

# aspera

ASPERA_REGEXP=re.compile(r'^fasp://(?P<url>.*)$')

@retry_if_it_fails(RETRY_TIME)
def fasp_get(source, destination):
    """Aspera downloader: download source expressed as fasp://user@host/path_to_file
    (NB usually with ENA, user is era-fasp) 
    to destination - a local file path"""
    log.info(f'Aspera downloading {source} to {destination}')
    if not docker_available:
        raise FetchError('Cannot use fasp (aspera) without docker')

    if source.endswith('/'):
        raise FetchError('Directory fetching is not yet supported for fasp:// URI (aspera)')

    # we need destination to be expressed as a directory:
    # (but we remember that the target filename may be different from the source
    # filename)
    if destination.endswith('/'):
        target_filename = None
    else:
        destination, target_filename = os.path.split(destination)
    
    uri_match = ASPERA_REGEXP.match(source).groupdict()
    subprocess.run(f'docker run --rm -v {destination}:/output ibmcom/aspera-cli \
ascp -T --policy=high -l 300m -P33001 -m 30m -v \
-i /home/aspera/.aspera/cli/etc/asperaweb_id_dsa.openssh \
{uri_match["url"]} /output/', shell=True, check=True)
    
    # if the target filename is really different 
    if target_filename is not None:
        source_filename = uri_match['url'].split('/')[-1]
        if source_filename!=target_filename:
            shutil.move(os.path.join(destination, source_filename),
                os.path.join(destination,target_filename))

# run accessions containing fastq


FASTQ_RUN_REGEXP=re.compile(r'^run\+fastq://(?P<run_accession>[^/]*)/?$')

@retry_if_it_fails(RETRY_TIME)
def fastq_sra_get(run_accession, destination):
    """This subfunction of runacc_get is only called when EBI's ENA won't
    answer as NCBI's SRA while more complete is quite slow"""
    log.info(f'SRA get run+fastq://{run_accession} to {destination}')
    if not docker_available:
        raise FetchError('Cannot use SRA toolkit without docker')
    if not destination.endswith('/'):
            destination+='/'
    previous_fastq = glob.glob(os.path.join(destination, '*.fastq'))
    subprocess.run(f'docker run --rm -v {destination}:/destination ncbi/sra-tools \
        sh -c "cd /destination && prefetch {run_accession} && fasterq-dump -f --split-files {run_accession}"',
        shell=True,
        check=True)
    current_fastq = glob.glob(os.path.join(destination, '*.fastq'))
    fastqs = [fastq for fastq in current_fastq if fastq not in previous_fastq]
    log.info(f'Pigziping fastqs ({fastqs})')
    subprocess.run(['pigz','-f']+fastqs,
        check=True)
    if os.path.isdir(os.path.join(destination, run_accession)):
        shutil.rmtree(os.path.join(destination, run_accession))

def _my_fastq_download(method, url, md5, destination):
    """A small adhoc function to download and check a fastq through a ftp_url plus a md5"""
    filename = url.split('/')[-1]
    if method in ['fastq_ftp','submitted_ftp']:
        ftp_get('ftp://'+url, destination, __retry_number__=1)
    elif method in ['fastq_aspera','submitted_aspera']:
        fasp_get(f'fasp://era-fasp@{url}', destination, __retry_number__=1)
    else:
        raise FetchError(f'No such method: {method}')
    with open(os.path.join(destination,filename),"rb") as f:
        readable_hash = hashlib.md5(f.read()).hexdigest()
        if readable_hash!=md5:
            raise FetchError(f'{filename} md5: {readable_hash} does not match ENA md5 {md5}')


@retry_if_it_fails(PUBLIC_RETRY_TIME)
def fastq_run_get(source, destination):
    """Fetch some fastq associated to a run accession"""
    log.info(f'Run accession: uploading {source} to {destination}')
    if not destination.endswith('/'):
            destination+='/'
    uri_match = FASTQ_RUN_REGEXP.match(source).groupdict()
    query_try = RETRY_TIME+1
    while query_try>0:
        while query_try>0:
            try:
                run_query = requests.get(f"https://www.ebi.ac.uk/ena/portal/api/filereport?\
accession={uri_match['run_accession']}&result=read_run&fields=fastq_md5,fastq_aspera,\
fastq_ftp,sra_md5,sra_ftp&format=json&download=true&limit=0", timeout=30)
            except requests.Timeout:
                query_try -= 1
                continue
            break
        else:
            log.exception('EBI does not answer our query')
            return fastq_sra_get(uri_match['run_accession'], destination)
        if run_query.status_code==204:
            log.exception('This does not seem to be available on EBI')
            return fastq_sra_get(uri_match['run_accession'], destination)
        runs = run_query.json()
        if len(runs)==0:
            # it seems the new API of EBI tends to answer empty responses
            # hoping this will change in near future
            query_try -= 1
            continue
        else:
            run=runs[0]
            break
    else:
        log.exception('EBI does not answer our query')
        return fastq_sra_get(uri_match['run_accession'], destination)
        



    if 'fastq_ftp' or 'fastq_aspera' in run:
        ftp_md5s = run['fastq_md5'].split(';')

        for method in ['fastq_ftp', 'sra', 'fastq_aspera']:
            if method == 'sra':
                return fastq_sra_get(uri_match['run_accession'], destination, 
                    __retry_number__=1)
            elif method in run:
                if method in ['fastq_aspera','sra'] and not docker_available:
                    continue
                urls =  run[method].split(';')

                # preparing to download and check all fastqs
                download_threads = []
                for url,md5 in zip(urls, ftp_md5s):
                    download_thread=PropagatingThread(target=_my_fastq_download,
                                        args=(method, url, md5, destination))
                    download_thread.start()
                    download_threads.append(download_thread)

                # waiting for all FTP to complete
                try:
                    for download_thread in download_threads:
                        download_thread.join()
                    return None
                except:
                    log.exception(f'EBI failed with method {method}')
                    continue
        raise FetchError(f'Could not fetch {source}')
    else:
        log.exception('EBI response does not include a fastq_ftp field')
        return fastq_sra_get(uri_match['run_accession'], destination)


SUBMITTED_RUN_REGEXP=re.compile(r'^run\+submitted://(?P<run_accession>[^/]*)/?$')

@retry_if_it_fails(PUBLIC_RETRY_TIME)
def submitted_run_get(source, destination):
    """Fetch some fastq associated to a run accession"""
    log.info(f'Run accession: uploading {source} to {destination}')
    if not destination.endswith('/'):
            destination+='/'
    uri_match = SUBMITTED_RUN_REGEXP.match(source).groupdict()
    query_try = RETRY_TIME+1
    while query_try>0:
        try:
            run_query = requests.get(f"https://www.ebi.ac.uk/ena/portal/api/filereport?\
accession={uri_match['run_accession']}&result=read_run&fields=submitted_md5,submitted_aspera,\
submitted_ftp&format=json&download=true&limit=0", timeout=30)
        except requests.Timeout:
            query_try -= 1
            continue
        break
    else:
        raise FetchError('EBI does not answer our query')
    if run_query.status_code==204:
        raise FetchError('This does not seem to be available on EBI')
    run = run_query.json()[0]



    if 'submitted_ftp' or 'submitted_aspera' in run:
        ftp_md5s = run['submitted_md5'].split(';')

        for method in ['submitted_ftp', 'submitted_aspera']:
            if method in run:
                if method in ['submitted_aspera'] and not docker_available:
                    continue
                urls =  run[method].split(';')

                # preparing to download and check all fastqs
                download_threads = []
                for url,md5 in zip(urls, ftp_md5s):
                    download_thread=PropagatingThread(target=_my_fastq_download,
                                        args=(method, url, md5, destination))
                    download_thread.start()
                    download_threads.append(download_thread)

                # waiting for all FTP to complete
                try:
                    for download_thread in download_threads:
                        download_thread.join()
                    return None
                except:
                    log.exception(f'EBI failed with method {method}')
                    continue
        raise FetchError(f'Could not fetch {source}')
    else:
        raise FetchError(f'Could not fetch {source}')


# plain file

FILE_REGEXP=re.compile(r'^file://(?P<path>.*)$')

def file_get(source, destination):
    """A plain local copy from a file://... source to a local path
    really just some plain syntaxic sugar above shutil.copyfile"""
    log.info(f'FILE downloading {source} to {destination}')
    if source.endswith('/'):
        raise FetchError('Directory fetching is not yet supported for file:// URI')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FILE_REGEXP.match(source).groupdict()
    shutil.copyfile(uri_match['path'], destination)

def file_put(source, destination):
    """Same as above except that source is this time a plain local path
    and destination is in the form file://... As above just some plain
    syntaxic sugar above shutil.copyfile """
    log.info(f'FILE uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FILE_REGEXP.match(destination).groupdict()
    if uri_match:
        complete_path = uri_match['path']
        path=os.path.dirname(complete_path)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        shutil.copyfile(source, complete_path)
    else:
        raise FetchError(f"Local URL did not match file://<path> pattern {destination}")

def file_info(uri):
    """Same as above except that source is this time a plain local path
    and destination is in the form file://... As above just some plain
    syntaxic sugar above shutil.copyfile """
    log.info(f'FILE getting info for {uri}')
    uri_match = FILE_REGEXP.match(uri).groupdict()
    if uri_match:
        complete_path = uri_match['path']
        stat = os.stat(complete_path)
        return argparse.Namespace(size=stat.st_size, 
                                  creation_date=datetime.datetime.utcfromtimestamp(stat.st_ctime).replace(tzinfo=pytz.utc),
                                  modification_date=datetime.datetime.utcfromtimestamp(stat.st_mtime).replace(tzinfo=pytz.utc))
    else:
        raise FetchError(f"Local URL did not match file://<path> pattern {uri}")


def file_list(uri, no_rec=False):
    """List recursively the content of a local folder expressed as file://... 
    Not recursively if no_rec is True"""
    log.info(f'FILE getting info for {uri}')
    uri_match = FILE_REGEXP.match(uri).groupdict()
    if uri_match:
        answer = []
        complete_path = uri_match['path']
        if no_rec:
            for file in os.listdir(complete_path):
                complete_file = os.path.join(complete_path,file)
                if os.path.isdir(complete_file):
                    file+='/'
                    complete_file+='/'
                stat = os.stat(complete_file)
                answer.append(argparse.Namespace(name=f'file://{complete_file}',
                                rel_name=file,
                                size=stat.st_size, 
                                creation_date=datetime.datetime.utcfromtimestamp(stat.st_ctime).replace(tzinfo=pytz.utc),
                                modification_date=datetime.datetime.utcfromtimestamp(stat.st_mtime).replace(tzinfo=pytz.utc)))
        else:
            for dir_path,_,files in os.walk(complete_path):
                for file in files:
                    complete_file = os.path.join(dir_path,file)
                    stat = os.stat(complete_file)
                    answer.append(argparse.Namespace(name=f'file://{complete_file}',
                                    rel_name=os.path.relpath(complete_file,complete_path),
                                    size=stat.st_size, 
                                    creation_date=datetime.datetime.utcfromtimestamp(stat.st_ctime).replace(tzinfo=pytz.utc),
                                    modification_date=datetime.datetime.utcfromtimestamp(stat.st_mtime).replace(tzinfo=pytz.utc)))
        return answer
    else:
        raise FetchError(f"Local URL did not match file://<path> pattern {uri}")

def file_delete(uri):
    """Delete a local file expressed as file://... """
    log.info(f'FILE deliting {uri}')
    uri_match = FILE_REGEXP.match(uri).groupdict()
    if uri_match:
        if os.path.exists(uri_match['path']):
            os.remove(uri_match['path'])
        else:
            raise FetchError(f"Local file {uri} does not seem to exist")
    else:
        raise FetchError(f"Local URL did not match file://<path> pattern {uri}")

# generic wrapper

GENERIC_REGEXP=re.compile(r'^(?P<proto>[a-z0-9+]*)://(?P<resource>[^|]*)(\|(?P<action>.*))?$')

def get(uri, destination):
    """General downloader source should start with s3://... or ftp://...
    (source should not end with slash unless you know what you are doing if 
    destination ends with slash it will be completed with source end item)
    file://... is also supported but be careful that it is local to worker.
    """
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        source = f"{m['proto']}://{m['resource']}"
        complete_destination = complete_if_ends_with_slash(source, destination)
        if uri.endswith('/'):
            complete_destination=complete_destination[:-1]
            complete_destination_folder=complete_destination            
        else:
            complete_destination_folder = '/'.join(complete_destination.split('/')[:-1])
        if not os.path.exists(complete_destination_folder):
            os.makedirs(complete_destination_folder, exist_ok=True)
        
        if m['proto']=='s3':
            s3_get(source, destination)
        elif m['proto']=='azure':
            AzureClient().get(source, destination) 
        elif m['proto']=='ftp':
            ftp_get(source, destination)
        elif m['proto']=='file':
            file_get(source, destination) 
        elif m['proto']=='fasp':
            fasp_get(source, destination)
        elif m['proto']=='run+fastq':
            fastq_run_get(source, destination)       
        elif m['proto']=='run+submitted':
            submitted_run_get(source, destination)
        else:
            raise FetchError(f"This URI protocol is not supported: {m['proto']}")
        
        if m['action']=='gunzip':
            gunzip(complete_destination)
        elif m['action']=='untar':
            untar(complete_destination)
        elif m['action'] not in ['',None]:
            raise FetchError(f"Unsupported action: {m['action']}")
    else:
        raise FetchError(f'This URI is malformed: {uri}')

def put(source, uri):
    """General uploader destination should start with s3://.... or ftp://...
    (only anonymous ftp is implemented so put is unlikely to work with ftp)
    (source should not end with slash unless you know what you are doing if 
    destination ends with slash it will be completed with source end item)
    file://... is also supported but be careful that it is local to worker."""
    m = GENERIC_REGEXP.match(uri)
    if source.endswith('/'):
        raise FetchError('Recursive put is not yet supported, invalid source: '+source)
    if m:
        m = m.groupdict()
        if m['action']:
            raise FetchError(f'Action are unsupported when putting in {uri}')
        destination = f"{m['proto']}://{m['resource']}"
        destination = complete_if_ends_with_slash(source, destination)
        if m['proto']=='s3':
            s3_put(source, destination)
        elif m['proto']=='azure':
            AzureClient().put(source, destination) 
        elif m['proto']=='ftp':
            ftp_put(source, destination)
        elif m['proto']=='file':
            file_put(source, destination)        
        else:
            raise FetchError(f"This URI proto is not supported: {m['proto']}")
    else:
        raise FetchError(f'This URI is malformed: {uri}')


def delete(uri):
    """General deleter for a URI."""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        if m['action']:
            raise FetchError(f'Action are unsupported when deleting a {uri}')
        if m['proto']=='s3':
            s3_delete(uri)
        elif m['proto']=='azure':
            AzureClient().delete(uri) 
        elif m['proto']=='file':
            file_delete(uri)        
        else:
            raise FetchError(f"This URI proto is not supported: {m['proto']}")
    else:
        raise FetchError(f'This URI is malformed: {uri}')


def check_uri(uri):
    """A small utility to check URI: return True if URI is valid, raise an exception otherwise"""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        if m['proto'] not in ['ftp','file','s3','azure','run+fastq','run+submitted']:
            raise FetchError(f"Unsupported protocol {m['proto']} in URI {uri}")
    else:
        raise FetchError(f"Malformed URI : {uri}")
    return True

def get_file_uri(uri):
    """A small utility to check if this is a file URI (starts with file://... or is a path)
    returns a path or None (if not a file URI)"""
    if uri.startswith('file://'):
        return uri[8:]
    elif ':' not in uri:
        return uri
    else:
        return None

def info(uri):
    """Return an info object from a URI, which should contains at least creation_date, modification_date and size
    WARNING: with s3 or ftp URI, creation_date and modification_date are the same."""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        source = f"{m['proto']}://{m['resource']}"
        if m['proto']=='s3':
            return s3_info(source)
        elif m['proto']=='azure':
            return AzureClient().info(source) 
        elif m['proto']=='ftp':
            return ftp_info(source)
        elif m['proto']=='file':
            return file_info(source) 
        #elif m['proto']=='fasp':
        #    fasp_get(source, destination)
        #elif m['proto']=='run+fastq':
        #    fastq_run_get(source, destination)       
        else:
            raise FetchError(f"This URI protocol is not supported: {m['proto']}")

def list_content(uri, no_rec=False):
    """Return the recursive listing of folder specified as a URI
    each item of the list should contains at least name (complete URI), rel_name (the name of the object relative to the provided uri), creation_date, modification_date and size
    if no_rec is True, then the listing is non recursive -> MEANS SOME "FOLDERS" WILL APPEAR IN LISTING, eventually with None attribute (in dates)
    WARNING: with s3 or ftp URI, creation_date and modification_date are the same.
    WARNING: this does not respect the pseudo-relativeness of s3/azure which is relative to the bucket/container, here the relativeness is to the folder URI"""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        source = f"{m['proto']}://{m['resource']}"
        if m['proto']=='s3':
            return s3_list(source, no_rec=no_rec)
        elif m['proto']=='azure':
            return AzureClient().list(source, no_rec=no_rec) 
        elif m['proto']=='ftp':
            return ftp_list(source, no_rec=no_rec)
        elif m['proto']=='file':
            return file_list(source, no_rec=no_rec) 
        #elif m['proto']=='fasp':
        #    fasp_get(source, destination)
        #elif m['proto']=='run+fastq':
        #    fastq_run_get(source, destination)       
        else:
            raise FetchError(f"This URI protocol is not supported: {m['proto']}")


def sync(uri1, uri2, include=None, process=MAX_PARALLEL_SYNC):
    """Sync two URI, one must be local
    Identity of the files is assessed only with name and size
    """
    local_uri = get_file_uri(uri1)
    if local_uri is not None:
        command = put
        remote = uri2
        try:
            check_uri(uri2)
        except FetchError:
            local_uri=get_file_uri(uri2)
            if local_uri is None:
                raise FetchError(f'uri2 appears to be neither locale nor prope: {uri2}')
            else:
                command=get
                remote=uri1 if ':' in uri1 else f'file://{uri1}'
    else:
        local_uri = get_file_uri(uri2)
        if local_uri is None:
            raise FetchError('Neither URI seems to be local, unsupported yet')
        command=get
        check_uri(uri1)
        remote=uri1
    full_local_uri=f"file://{local_uri}"
    if not os.path.exists(local_uri):
        os.makedirs(local_uri, exist_ok=True)

    local_list = dict([(item.rel_name,item) for item in list_content(full_local_uri)])
    remote_list = dict([(item.rel_name,item) for item in list_content(remote)])

    if command==get:
        source_list=remote_list
        dest_list=local_list
        source_uri=remote
        dest_uri=local_uri
    else:
        source_list=local_list
        dest_list=remote_list
        source_uri=local_uri
        dest_uri=remote

    jobs = {}
    with concurrent.futures.ProcessPoolExecutor(max_workers=process) as executor:
        for item_name, item in source_list.items():
            if item_name not in dest_list or \
                    dest_list[item_name].size!=item.size:
                if include:
                    for inc in include:
                        if fnmatch(item_name, inc):
                            break
                    else:
                        continue
                log.warning(f'Copying {os.path.join(source_uri, item_name)} to {os.path.join(dest_uri, item_name)}')
                jobs[executor.submit(command, 
                                os.path.join(source_uri, item_name),
                                os.path.join(dest_uri, item_name))]=item_name
    
    failed = False
    for job in concurrent.futures.as_completed(jobs):
        item_name = jobs[job]
        if job.exception() is None:
            log.info(f'Done for {item_name}: {job.result()}')
        else:
            log.error(f'Could not download {item_name}')
            log.exception(job.exception())
            failed = True
    
    if failed:
        raise FetchError('At least some objects could not be synchronized')

def recursive_delete(uri, include=None, dryrun=False):
    """Works the same way than sync, recursively deleting some objects from uri"""
    jobs = {}
    with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_PARALLEL_SYNC) as executor:
        for item in list_content(uri):
            item_name=item.rel_name.split('/')[-1]
            if include:
                for inc in include:
                    if fnmatch(item_name, inc):
                        break
                else:
                    continue
            log.warning(f'Deleting {item.name}')
            if not dryrun:
                jobs[executor.submit(delete, 
                            item.name)]=item.name
    
    failed = False
    for job in concurrent.futures.as_completed(jobs):
        item_name = jobs[job]
        if job.exception() is None:
            log.info(f'Done for {item_name}: {job.result()}')
        else:
            log.error(f'Could not delete {item_name}')
            log.exception(job.exception())
            failed = True

    if failed:
        raise FetchError('At least some objects could not be deleted')


def main():
    parser = argparse.ArgumentParser(
                    prog = 'scitq.fetch module utility mode',
                    description = '''Enable direct download or upload without python code.
Takes a command with one or two URI (generalised URL, including file://... and s3://... or ftp://...) 
one of them must be a file URI (starts with file://... or be a simple path)''')
    parser.add_argument('-v','--verbose',action='store_true',help='Turn log level to info')
    parser.add_argument('-p','--process',type=int,default=MAX_PARALLEL_SYNC,
                        help=f'Number of parallel process (default to {MAX_PARALLEL_SYNC})')
    subparser=parser.add_subparsers(help='sub-command help',dest='command')

    get_parser = subparser.add_parser('copy', help='Copy some file or folder to some folder (one of them must be local)')
    get_parser.add_argument('source_uri', type=str, help='the uri (can be a local file or a remote URI)')
    get_parser.add_argument('destination_uri', type=str, nargs='?',
                        help='the destination uri (same as above, default to ., means download locally)', default=os.getcwd())
    
    list_parser = subparser.add_parser('list', help='List the content of a remote folder (outputs some absolute URI)')
    list_parser.add_argument('--not-recursive', action="store_true", help="Do not be recursive")
    list_parser.add_argument('uri', type=str, help='the remote folder uri')
    
    rlist_parser = subparser.add_parser('rlist', help='List the content of a remote folder (outputs some relative path to the URI)')
    rlist_parser.add_argument('--not-recursive', action="store_true", help="Do not be recursive")
    rlist_parser.add_argument('uri', type=str, help='the remote folder uri')

    sync_parser = subparser.add_parser('sync', help='Sync some file or folder to some folder (one of them must be local) (identity is checked using name and size)')
    sync_parser.add_argument('source_uri', type=str, help='the uri (can be a local file or a remote URI)')
    sync_parser.add_argument('destination_uri', type=str,
                        help='the destination uri (same as above, default to ., means download locally)', default=os.getcwd())
    sync_parser.add_argument('--include',action='append',type=str,
                        help="A pattern that should be included (only those will be synced) (can be specified several times)")
    

    delete_parser = subparser.add_parser('delete', help='Delete some file or folder expressed as a URI')
    delete_parser.add_argument('uri', type=str, help='the uri (can be a local file or a remote URI)')
    delete_parser.add_argument('--include',action='append',type=str,
                        help="A pattern that should be included (only those will be synced) (can be specified several times)")
    delete_parser.add_argument('--dryrun',action='store_true',
                        help="Do not really delete, just print what it would delete")

    args = parser.parse_args()

    if args.verbose:
        log.basicConfig(level=log.INFO)

    if args.command=='copy':
        candidate_source = get_file_uri(args.source_uri)
        if candidate_source is not None:
            check_uri(args.destination_uri)
            put(candidate_source, args.destination_uri)
        else:
            candidate_destination = get_file_uri(args.destination_uri)
            if candidate_destination is not None:
                check_uri(args.source_uri)
                get(args.source_uri, candidate_destination)
            else:
                raise RuntimeError('Both source_uri and destination_uri seem non file URI: operation unsupported')
    elif args.command=='list':
        if ':' not in args.uri:
            uri=f'file://{args.uri}'
        else:
            check_uri(args.uri)
            uri=args.uri
        headers=['name','creation_date','modification_date','size']
        print(tabulate(
            [[ if_is_not_None(getattr(item,attribute),'-') for attribute in headers  ] for item in list_content(uri, no_rec=args.not_recursive)],
            headers=headers,
            tablefmt='plain'
        ))
    elif args.command=='rlist':
        if ':' not in args.uri:
            uri=f'file://{args.uri}'
        else:
            check_uri(args.uri)
            uri=args.uri
        headers=['rel_name','creation_date','modification_date','size']
        print(tabulate(
            [[ if_is_not_None(getattr(item,attribute),'-') for attribute in headers  ] for item in list_content(uri, no_rec=args.not_recursive)],
            headers=headers,
            tablefmt='plain'
        ))
    elif args.command=='sync':
        sync(args.source_uri, args.destination_uri, include=args.include, process=args.process)
    elif args.command=='delete':
        recursive_delete(args.uri, include=args.include, dryrun=args.dryrun)

if __name__=='__main__':
    main()